# Copyright 2025-2026 Dimensional Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
import asyncio
from collections.abc import Callable
from dataclasses import dataclass
from functools import partial
import sys
import threading
from typing import (
    TYPE_CHECKING,
    Any,
    get_args,
    get_origin,
    get_type_hints,
    overload,
)

if TYPE_CHECKING:
    from dimos.core.introspection.module import ModuleInfo

from dask.distributed import Actor, get_worker
from reactivex.disposable import CompositeDisposable
from typing_extensions import TypeVar

from dimos.core import colors
from dimos.core.core import T, rpc
from dimos.core.introspection.module import extract_module_info, render_module_io
from dimos.core.resource import Resource
from dimos.core.rpc_client import RpcCall
from dimos.core.stream import In, Out, RemoteIn, RemoteOut, Transport
from dimos.protocol.rpc import LCMRPC, RPCSpec
from dimos.protocol.service import Configurable  # type: ignore[attr-defined]
from dimos.protocol.skill.skill import SkillContainer
from dimos.protocol.tf import LCMTF, TFSpec
from dimos.utils.generic import classproperty


def get_loop() -> tuple[asyncio.AbstractEventLoop, threading.Thread | None]:
    # we are actually instantiating a new loop here
    # to not interfere with an existing dask loop

    # try:
    #     # here we attempt to figure out if we are running on a dask worker
    #     # if so we use the dask worker _loop as ours,
    #     # and we register our RPC server
    #     worker = get_worker()
    #     if worker.loop:
    #         print("using dask worker loop")
    #         return worker.loop.asyncio_loop

    # except ValueError:
    #     ...

    try:
        running_loop = asyncio.get_running_loop()
        return running_loop, None
    except RuntimeError:
        loop = asyncio.new_event_loop()
        asyncio.set_event_loop(loop)

        thr = threading.Thread(target=loop.run_forever, daemon=True)
        thr.start()
        return loop, thr


@dataclass
class ModuleConfig:
    rpc_transport: type[RPCSpec] = LCMRPC
    tf_transport: type[TFSpec] = LCMTF
    frame_id_prefix: str | None = None
    frame_id: str | None = None


ModuleConfigT = TypeVar("ModuleConfigT", bound=ModuleConfig, default=ModuleConfig)


class ModuleBase(Configurable[ModuleConfigT], SkillContainer, Resource):
    _rpc: RPCSpec | None = None
    _tf: TFSpec | None = None
    _loop: asyncio.AbstractEventLoop | None = None
    _loop_thread: threading.Thread | None
    _disposables: CompositeDisposable
    _bound_rpc_calls: dict[str, RpcCall] = {}

    rpc_calls: list[str] = []

    default_config: type[ModuleConfigT] = ModuleConfig  # type: ignore[assignment]

    def __init__(self, *args, **kwargs) -> None:  # type: ignore[no-untyped-def]
        super().__init__(*args, **kwargs)
        self._loop, self._loop_thread = get_loop()
        self._disposables = CompositeDisposable()
        # we can completely override comms protocols if we want
        try:
            # here we attempt to figure out if we are running on a dask worker
            # if so we use the dask worker _loop as ours,
            # and we register our RPC server
            self.rpc = self.config.rpc_transport()
            self.rpc.serve_module_rpc(self)
            self.rpc.start()  # type: ignore[attr-defined]
        except ValueError:
            ...

    @property
    def frame_id(self) -> str:
        base = self.config.frame_id or self.__class__.__name__
        if self.config.frame_id_prefix:
            return f"{self.config.frame_id_prefix}/{base}"
        return base

    @rpc
    def start(self) -> None:
        pass

    @rpc
    def stop(self) -> None:
        self._close_module()
        super().stop()

    def _close_module(self) -> None:
        self._close_rpc()
        if hasattr(self, "_loop") and self._loop_thread:
            if self._loop_thread.is_alive():
                self._loop.call_soon_threadsafe(self._loop.stop)  # type: ignore[union-attr]
                self._loop_thread.join(timeout=2)
            self._loop = None
            self._loop_thread = None
        if hasattr(self, "_tf") and self._tf is not None:
            self._tf.stop()
            self._tf = None
        if hasattr(self, "_disposables"):
            self._disposables.dispose()

    def _close_rpc(self) -> None:
        # Using hasattr is needed because SkillCoordinator skips ModuleBase.__init__ and self.rpc is never set.
        if hasattr(self, "rpc") and self.rpc:
            self.rpc.stop()  # type: ignore[attr-defined]
            self.rpc = None  # type: ignore[assignment]

    def __getstate__(self):  # type: ignore[no-untyped-def]
        """Exclude unpicklable runtime attributes when serializing."""
        state = self.__dict__.copy()
        # Remove unpicklable attributes
        state.pop("_disposables", None)
        state.pop("_loop", None)
        state.pop("_loop_thread", None)
        state.pop("_rpc", None)
        state.pop("_tf", None)
        return state

    def __setstate__(self, state) -> None:  # type: ignore[no-untyped-def]
        """Restore object from pickled state."""
        self.__dict__.update(state)
        # Reinitialize runtime attributes
        self._disposables = CompositeDisposable()
        self._loop = None
        self._loop_thread = None
        self._rpc = None
        self._tf = None

    @property
    def tf(self):  # type: ignore[no-untyped-def]
        if self._tf is None:
            # self._tf = self.config.tf_transport()
            self._tf = LCMTF()
        return self._tf

    @tf.setter
    def tf(self, value) -> None:  # type: ignore[no-untyped-def]
        import warnings

        warnings.warn(
            "tf is available on all modules. Call self.tf.start() to activate tf functionality. No need to assign it",
            UserWarning,
            stacklevel=2,
        )

    @property
    def outputs(self) -> dict[str, Out]:  # type: ignore[type-arg]
        return {
            name: s
            for name, s in self.__dict__.items()
            if isinstance(s, Out) and not name.startswith("_")
        }

    @property
    def inputs(self) -> dict[str, In]:  # type: ignore[type-arg]
        return {
            name: s
            for name, s in self.__dict__.items()
            if isinstance(s, In) and not name.startswith("_")
        }

    @classproperty
    def rpcs(self) -> dict[str, Callable[..., Any]]:
        return {
            name: getattr(self, name)
            for name in dir(self)
            if not name.startswith("_")
            and name != "rpcs"  # Exclude the rpcs property itself to prevent recursion
            and callable(getattr(self, name, None))
            and hasattr(getattr(self, name), "__rpc__")
        }

    @rpc
    def _io_instance(self, color: bool = True) -> str:
        """Instance-level io() - shows actual running streams."""
        return render_module_io(
            name=self.__class__.__name__,
            inputs=self.inputs,
            outputs=self.outputs,
            rpcs=self.rpcs,
            color=color,
        )

    @classmethod
    def _io_class(cls, color: bool = True) -> str:
        """Class-level io() - shows declared stream types from annotations."""
        hints = get_type_hints(cls)

        _yellow = colors.yellow if color else (lambda x: x)
        _green = colors.green if color else (lambda x: x)

        def is_stream(hint: type, stream_type: type) -> bool:
            origin = get_origin(hint)
            if origin is stream_type:
                return True
            if isinstance(hint, type) and issubclass(hint, stream_type):
                return True
            return False

        def format_stream(name: str, hint: type) -> str:
            args = get_args(hint)
            type_name = args[0].__name__ if args else "?"
            return f"{_yellow(name)}: {_green(type_name)}"

        inputs = {
            name: format_stream(name, hint) for name, hint in hints.items() if is_stream(hint, In)
        }
        outputs = {
            name: format_stream(name, hint) for name, hint in hints.items() if is_stream(hint, Out)
        }

        return render_module_io(
            name=cls.__name__,
            inputs=inputs,
            outputs=outputs,
            rpcs=cls.rpcs,
            color=color,
        )

    class _io_descriptor:
        """Descriptor that makes io() work on both class and instance."""

        def __get__(
            self, obj: "ModuleBase | None", objtype: type["ModuleBase"]
        ) -> Callable[[bool], str]:
            if obj is None:
                return objtype._io_class
            return obj._io_instance

    io = _io_descriptor()

    @classmethod
    def _module_info_class(cls) -> "ModuleInfo":
        """Class-level module_info() - returns ModuleInfo from annotations."""

        hints = get_type_hints(cls)

        def is_stream(hint: type, stream_type: type) -> bool:
            origin = get_origin(hint)
            if origin is stream_type:
                return True
            if isinstance(hint, type) and issubclass(hint, stream_type):
                return True
            return False

        def format_stream(name: str, hint: type) -> str:
            args = get_args(hint)
            type_name = args[0].__name__ if args else "?"
            return f"{name}: {type_name}"

        inputs = {
            name: format_stream(name, hint) for name, hint in hints.items() if is_stream(hint, In)
        }
        outputs = {
            name: format_stream(name, hint) for name, hint in hints.items() if is_stream(hint, Out)
        }

        return extract_module_info(
            name=cls.__name__,
            inputs=inputs,
            outputs=outputs,
            rpcs=cls.rpcs,
        )

    class _module_info_descriptor:
        """Descriptor that makes module_info() work on both class and instance."""

        def __get__(
            self, obj: "ModuleBase | None", objtype: type["ModuleBase"]
        ) -> Callable[[], "ModuleInfo"]:
            if obj is None:
                return objtype._module_info_class
            # For instances, extract from actual streams
            return lambda: extract_module_info(
                name=obj.__class__.__name__,
                inputs=obj.inputs,
                outputs=obj.outputs,
                rpcs=obj.rpcs,
            )

    module_info = _module_info_descriptor()

    @classproperty
    def blueprint(self):  # type: ignore[no-untyped-def]
        # Here to prevent circular imports.
        from dimos.core.blueprints import create_module_blueprint

        return partial(create_module_blueprint, self)  # type: ignore[arg-type]

    @rpc
    def get_rpc_method_names(self) -> list[str]:
        return self.rpc_calls

    @rpc
    def set_rpc_method(self, method: str, callable: RpcCall) -> None:
        callable.set_rpc(self.rpc)  # type: ignore[arg-type]
        self._bound_rpc_calls[method] = callable

    @overload
    def get_rpc_calls(self, method: str) -> RpcCall: ...

    @overload
    def get_rpc_calls(self, method1: str, method2: str, *methods: str) -> tuple[RpcCall, ...]: ...

    def get_rpc_calls(self, *methods: str) -> RpcCall | tuple[RpcCall, ...]:  # type: ignore[misc]
        missing = [m for m in methods if m not in self._bound_rpc_calls]
        if missing:
            raise ValueError(
                f"RPC methods not found. Class: {self.__class__.__name__}, RPC methods: {', '.join(missing)}"
            )
        result = tuple(self._bound_rpc_calls[m] for m in methods)
        return result[0] if len(result) == 1 else result


class DaskModule(ModuleBase[ModuleConfigT]):
    ref: Actor
    worker: int

    def __init_subclass__(cls, **kwargs: Any) -> None:
        """Set class-level None attributes for In/Out type annotations.

        This is needed because Dask's Actor proxy looks up attributes on the class
        (not instance) when proxying attribute access. Without class-level attributes,
        the proxy would fail with AttributeError even though the instance has the attrs.
        """
        super().__init_subclass__(**kwargs)

        # Get type hints for this class only (not inherited ones).
        globalns = {}
        for c in cls.__mro__:
            if c.__module__ in sys.modules:
                globalns.update(sys.modules[c.__module__].__dict__)

        try:
            hints = get_type_hints(cls, globalns=globalns, include_extras=True)
        except (NameError, AttributeError, TypeError):
            hints = {}

        for name, ann in hints.items():
            origin = get_origin(ann)
            if origin in (In, Out):
                # Set class-level attribute if not already set.
                if not hasattr(cls, name) or getattr(cls, name) is None:
                    setattr(cls, name, None)

    def __init__(self, *args, **kwargs) -> None:  # type: ignore[no-untyped-def]
        self.ref = None  # type: ignore[assignment]

        # Get type hints with proper namespace resolution for subclasses
        # Collect namespaces from all classes in the MRO chain
        import sys

        globalns = {}
        for cls in self.__class__.__mro__:
            if cls.__module__ in sys.modules:
                globalns.update(sys.modules[cls.__module__].__dict__)

        try:
            hints = get_type_hints(self.__class__, globalns=globalns, include_extras=True)
        except (NameError, AttributeError, TypeError):
            # If we still can't resolve hints, skip type hint processing
            # This can happen with complex forward references
            hints = {}

        for name, ann in hints.items():
            origin = get_origin(ann)
            if origin is Out:
                inner, *_ = get_args(ann) or (Any,)
                stream = Out(inner, name, self)  # type: ignore[var-annotated]
                setattr(self, name, stream)
            elif origin is In:
                inner, *_ = get_args(ann) or (Any,)
                stream = In(inner, name, self)  # type: ignore[assignment]
                setattr(self, name, stream)
        super().__init__(*args, **kwargs)

    def set_ref(self, ref) -> int:  # type: ignore[no-untyped-def]
        worker = get_worker()
        self.ref = ref
        self.worker = worker.name
        return worker.name  # type: ignore[no-any-return]

    def __str__(self) -> str:
        return f"{self.__class__.__name__}"

    @rpc
    def set_transport(self, stream_name: str, transport: Transport) -> bool:  # type: ignore[type-arg]
        stream = getattr(self, stream_name, None)
        if not stream:
            raise ValueError(f"{stream_name} not found in {self.__class__.__name__}")

        if not isinstance(stream, Out) and not isinstance(stream, In):
            raise TypeError(f"Output {stream_name} is not a valid stream")

        stream._transport = transport
        return True

    # called from remote
    def connect_stream(self, input_name: str, remote_stream: RemoteOut[T]):  # type: ignore[no-untyped-def]
        input_stream = getattr(self, input_name, None)
        if not input_stream:
            raise ValueError(f"{input_name} not found in {self.__class__.__name__}")
        if not isinstance(input_stream, In):
            raise TypeError(f"Input {input_name} is not a valid stream")
        input_stream.connection = remote_stream

    def dask_receive_msg(self, input_name: str, msg: Any) -> None:
        getattr(self, input_name).transport.dask_receive_msg(msg)

    def dask_register_subscriber(self, output_name: str, subscriber: RemoteIn[T]) -> None:
        getattr(self, output_name).transport.dask_register_subscriber(subscriber)


# global setting
Module = DaskModule



# Copyright 2025-2026 Dimensional Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from __future__ import annotations

import importlib
import re
import time

import typer

from dimos.core.transport import LCMTransport, pLCMTransport
from dimos.protocol.pubsub.lcmpubsub import LCMPubSubBase

_modules_to_try = [
    "dimos.msgs.geometry_msgs",
    "dimos.msgs.nav_msgs",
    "dimos.msgs.sensor_msgs",
    "dimos.msgs.std_msgs",
    "dimos.msgs.vision_msgs",
    "dimos.msgs.foxglove_msgs",
    "dimos.msgs.tf2_msgs",
]


def _resolve_type(type_name: str) -> type:
    for module_name in _modules_to_try:
        try:
            module = importlib.import_module(module_name)
            if hasattr(module, type_name):
                return getattr(module, type_name)  # type: ignore[no-any-return]
        except ImportError:
            continue

    raise ValueError(f"Could not find type '{type_name}' in any known message modules")


def topic_echo(topic: str, type_name: str | None) -> None:
    # Explicit mode (legacy): unchanged.
    if type_name is not None:
        msg_type = _resolve_type(type_name)
        use_pickled = getattr(msg_type, "lcm_encode", None) is None
        transport: pLCMTransport[object] | LCMTransport[object] = (
            pLCMTransport(topic) if use_pickled else LCMTransport(topic, msg_type)
        )

        def _on_message(msg: object) -> None:
            print(msg)

        transport.subscribe(_on_message)
        typer.echo(f"Listening on {topic} for {type_name} messages... (Ctrl+C to stop)")
        try:
            while True:
                time.sleep(0.1)
        except KeyboardInterrupt:
            typer.echo("\nStopped.")
        return

    # Inferred typed mode: listen on /topic#pkg.Msg and decode from the msg_name suffix.
    bus = LCMPubSubBase(autoconf=True)
    bus.start()  # starts threaded handle loop

    typed_pattern = rf"^{re.escape(topic)}#.*"

    def on_msg(channel: str, data: bytes) -> None:
        _, msg_name = channel.split("#", 1)  # e.g. "nav_msgs.Odometry"
        pkg, cls_name = msg_name.split(".", 1)  # "nav_msgs", "Odometry"
        module = importlib.import_module(f"dimos.msgs.{pkg}")
        cls = getattr(module, cls_name)
        print(cls.lcm_decode(data))

    assert bus.l is not None
    bus.l.subscribe(typed_pattern, on_msg)

    typer.echo(
        f"Listening on {topic} (inferring from typed LCM channels like '{topic}#pkg.Msg')... "
        "(Ctrl+C to stop)"
    )

    try:
        while True:
            time.sleep(0.1)
    except KeyboardInterrupt:
        bus.stop()
        typer.echo("\nStopped.")


def topic_send(topic: str, message_expr: str) -> None:
    eval_context: dict[str, object] = {}
    modules_to_import = [
        "dimos.msgs.geometry_msgs",
        "dimos.msgs.nav_msgs",
        "dimos.msgs.sensor_msgs",
        "dimos.msgs.std_msgs",
        "dimos.msgs.vision_msgs",
        "dimos.msgs.foxglove_msgs",
        "dimos.msgs.tf2_msgs",
    ]

    for module_name in modules_to_import:
        try:
            module = importlib.import_module(module_name)
            for name in getattr(module, "__all__", dir(module)):
                if not name.startswith("_"):
                    obj = getattr(module, name, None)
                    if obj is not None:
                        eval_context[name] = obj
        except ImportError:
            continue

    try:
        message = eval(message_expr, eval_context)
    except Exception as e:
        typer.echo(f"Error parsing message: {e}", err=True)
        raise typer.Exit(1)

    msg_type = type(message)
    use_pickled = getattr(msg_type, "lcm_encode", None) is None
    transport: pLCMTransport[object] | LCMTransport[object] = (
        pLCMTransport(topic) if use_pickled else LCMTransport(topic, msg_type)
    )

    transport.broadcast(None, message)
    typer.echo(f"Sent to {topic}: {message}")



# Copyright 2025-2026 Dimensional Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

# Copyright 2025-2026 Dimensional Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     [http://www.apache.org/licenses/LICENSE-2.0](http://www.apache.org/licenses/LICENSE-2.0)
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""Module for manipulation history tracking and search."""

from dataclasses import dataclass, field
from datetime import datetime
import json
import os
import pickle
import time
from typing import Any

from dimos.types.manipulation import (
    ManipulationTask,
)
from dimos.utils.logging_config import setup_logger

logger = setup_logger()


@dataclass
class ManipulationHistoryEntry:
    """An entry in the manipulation history.

    Attributes:
        task: The manipulation task executed
        timestamp: When the manipulation was performed
        result: Result of the manipulation (success/failure)
        manipulation_response: Response from the motion planner/manipulation executor
    """

    task: ManipulationTask
    timestamp: float = field(default_factory=time.time)
    result: dict[str, Any] = field(default_factory=dict)
    manipulation_response: str | None = (
        None  # Any elaborative response from the motion planner / manipulation executor
    )

    def __str__(self) -> str:
        status = self.result.get("status", "unknown")
        return f"ManipulationHistoryEntry(task='{self.task.description}', status={status}, time={datetime.fromtimestamp(self.timestamp).strftime('%H:%M:%S')})"


class ManipulationHistory:
    """A simplified, dictionary-based storage for manipulation history.

    This class provides an efficient way to store and query manipulation tasks,
    focusing on quick lookups and flexible search capabilities.
    """

    def __init__(self, output_dir: str | None = None, new_memory: bool = False) -> None:
        """Initialize a new manipulation history.

        Args:
            output_dir: Directory to save history to
            new_memory: If True, creates a new memory instead of loading existing one
        """
        self._history: list[ManipulationHistoryEntry] = []
        self._output_dir = output_dir

        if output_dir and not new_memory:
            self.load_from_dir(output_dir)
        elif output_dir:
            os.makedirs(output_dir, exist_ok=True)
            logger.info(f"Created new manipulation history at {output_dir}")

    def __len__(self) -> int:
        """Return the number of entries in the history."""
        return len(self._history)

    def __str__(self) -> str:
        """Return a string representation of the history."""
        if not self._history:
            return "ManipulationHistory(empty)"

        return (
            f"ManipulationHistory(entries={len(self._history)}, "
            f"time_range={datetime.fromtimestamp(self._history[0].timestamp).strftime('%Y-%m-%d %H:%M:%S')} to "
            f"{datetime.fromtimestamp(self._history[-1].timestamp).strftime('%Y-%m-%d %H:%M:%S')})"
        )

    def clear(self) -> None:
        """Clear all entries from the history."""
        self._history.clear()
        logger.info("Cleared manipulation history")

        if self._output_dir:
            self.save_history()

    def add_entry(self, entry: ManipulationHistoryEntry) -> None:
        """Add an entry to the history.

        Args:
            entry: The entry to add
        """
        self._history.append(entry)
        self._history.sort(key=lambda e: e.timestamp)

        if self._output_dir:
            self.save_history()

    def save_history(self) -> None:
        """Save the history to the output directory."""
        if not self._output_dir:
            logger.warning("Cannot save history: no output directory specified")
            return

        os.makedirs(self._output_dir, exist_ok=True)
        history_path = os.path.join(self._output_dir, "manipulation_history.pickle")

        with open(history_path, "wb") as f:
            pickle.dump(self._history, f)

        logger.info(f"Saved manipulation history to {history_path}")

        # Also save a JSON representation for easier inspection
        json_path = os.path.join(self._output_dir, "manipulation_history.json")
        try:
            history_data = [
                {
                    "task": {
                        "description": entry.task.description,
                        "target_object": entry.task.target_object,
                        "target_point": entry.task.target_point,
                        "timestamp": entry.task.timestamp,
                        "task_id": entry.task.task_id,
                        "metadata": entry.task.metadata,
                    },
                    "result": entry.result,
                    "timestamp": entry.timestamp,
                    "manipulation_response": entry.manipulation_response,
                }
                for entry in self._history
            ]

            with open(json_path, "w") as f:
                json.dump(history_data, f, indent=2)

            logger.info(f"Saved JSON representation to {json_path}")
        except Exception as e:
            logger.error(f"Failed to save JSON representation: {e}")

    def load_from_dir(self, directory: str) -> None:
        """Load history from the specified directory.

        Args:
            directory: Directory to load history from
        """
        history_path = os.path.join(directory, "manipulation_history.pickle")

        if not os.path.exists(history_path):
            logger.warning(f"No history found at {history_path}")
            return

        try:
            with open(history_path, "rb") as f:
                self._history = pickle.load(f)

            logger.info(
                f"Loaded manipulation history from {history_path} with {len(self._history)} entries"
            )
        except Exception as e:
            logger.error(f"Failed to load history: {e}")

    def get_all_entries(self) -> list[ManipulationHistoryEntry]:
        """Get all entries in chronological order.

        Returns:
            List of all manipulation history entries
        """
        return self._history.copy()

    def get_entry_by_index(self, index: int) -> ManipulationHistoryEntry | None:
        """Get an entry by its index.

        Args:
            index: Index of the entry to retrieve

        Returns:
            The entry at the specified index or None if index is out of bounds
        """
        if 0 <= index < len(self._history):
            return self._history[index]
        return None

    def get_entries_by_timerange(
        self, start_time: float, end_time: float
    ) -> list[ManipulationHistoryEntry]:
        """Get entries within a specific time range.

        Args:
            start_time: Start time (UNIX timestamp)
            end_time: End time (UNIX timestamp)

        Returns:
            List of entries within the specified time range
        """
        return [entry for entry in self._history if start_time <= entry.timestamp <= end_time]

    def get_entries_by_object(self, object_name: str) -> list[ManipulationHistoryEntry]:
        """Get entries related to a specific object.

        Args:
            object_name: Name of the object to search for

        Returns:
            List of entries related to the specified object
        """
        return [entry for entry in self._history if entry.task.target_object == object_name]

    def create_task_entry(
        self,
        task: ManipulationTask,
        result: dict[str, Any] | None = None,
        agent_response: str | None = None,
    ) -> ManipulationHistoryEntry:
        """Create a new manipulation history entry.

        Args:
            task: The manipulation task
            result: Result of the manipulation
            agent_response: Response from the agent about this manipulation

        Returns:
            The created history entry
        """
        entry = ManipulationHistoryEntry(
            task=task, result=result or {}, manipulation_response=agent_response
        )
        self.add_entry(entry)
        return entry

    def search(self, **kwargs) -> list[ManipulationHistoryEntry]:  # type: ignore[no-untyped-def]
        """Flexible search method that can search by any field in ManipulationHistoryEntry using dot notation.

        This method supports dot notation to access nested fields. String values automatically use
        substring matching (contains), while all other types use exact matching.

        Examples:
        # Time-based searches:
        - search(**{"task.metadata.timestamp": ('>', start_time)}) - entries after start_time
        - search(**{"task.metadata.timestamp": ('>=', time - 1800)}) - entries in last 30 mins

        # Constraint searches:
        - search(**{"task.constraints.*.reference_point.x": 2.5}) - tasks with x=2.5 reference point
        - search(**{"task.constraints.*.end_angle.x": 90}) - tasks with 90-degree x rotation
        - search(**{"task.constraints.*.lock_x": True}) - tasks with x-axis translation locked

        # Object and result searches:
        - search(**{"task.metadata.objects.*.label": "cup"}) - tasks involving cups
        - search(**{"result.status": "success"}) - successful tasks
        - search(**{"result.error": "Collision"}) - tasks that had collisions

        Args:
            **kwargs: Key-value pairs for searching using dot notation for field paths.

        Returns:
            List of matching entries
        """
        if not kwargs:
            return self._history.copy()

        results = self._history.copy()

        for key, value in kwargs.items():
            # For all searches, automatically determine if we should use contains for strings
            results = [e for e in results if self._check_field_match(e, key, value)]

        return results

    def _check_field_match(self, entry, field_path, value) -> bool:  # type: ignore[no-untyped-def]
        """Check if a field matches the value, with special handling for strings, collections and comparisons.

        For string values, we automatically use substring matching (contains).
        For collections (returned by * path), we check if any element matches.
        For numeric values (like timestamps), supports >, <, >= and <= comparisons.
        For all other types, we use exact matching.

        Args:
            entry: The entry to check
            field_path: Dot-separated path to the field
            value: Value to match against. For comparisons, use tuples like:
                  ('>',  timestamp) - greater than
                  ('<',  timestamp) - less than
                  ('>=', timestamp) - greater or equal
                  ('<=', timestamp) - less or equal

        Returns:
            True if the field matches the value, False otherwise
        """
        try:
            field_value = self._get_value_by_path(entry, field_path)  # type: ignore[no-untyped-call]

            # Handle comparison operators for timestamps and numbers
            if isinstance(value, tuple) and len(value) == 2:
                op, compare_value = value
                if op == ">":
                    return field_value > compare_value  # type: ignore[no-any-return]
                elif op == "<":
                    return field_value < compare_value  # type: ignore[no-any-return]
                elif op == ">=":
                    return field_value >= compare_value  # type: ignore[no-any-return]
                elif op == "<=":
                    return field_value <= compare_value  # type: ignore[no-any-return]

            # Handle lists (from collection searches)
            if isinstance(field_value, list):
                for item in field_value:
                    # String values use contains matching
                    if isinstance(item, str) and isinstance(value, str):
                        if value in item:
                            return True
                    # All other types use exact matching
                    elif item == value:
                        return True
                return False

            # String values use contains matching
            elif isinstance(field_value, str) and isinstance(value, str):
                return value in field_value
            # All other types use exact matching
            else:
                return field_value == value  # type: ignore[no-any-return]

        except (AttributeError, KeyError):
            return False

    def _get_value_by_path(self, obj, path):  # type: ignore[no-untyped-def]
        """Get a value from an object using a dot-separated path.

        This method handles three special cases:
        1. Regular attribute access (obj.attr)
        2. Dictionary key access (dict[key])
        3. Collection search (dict.*.attr) - when * is used, it searches all values in the collection

        Args:
            obj: Object to get value from
            path: Dot-separated path to the field (e.g., "task.metadata.robot")

        Returns:
            Value at the specified path or list of values for collection searches

        Raises:
            AttributeError: If an attribute in the path doesn't exist
            KeyError: If a dictionary key in the path doesn't exist
        """
        current = obj
        parts = path.split(".")

        for i, part in enumerate(parts):
            # Collection search (*.attr) - search across all items in a collection
            if part == "*":
                # Get remaining path parts
                remaining_path = ".".join(parts[i + 1 :])

                # Handle different collection types
                if isinstance(current, dict):
                    items = current.values()
                    if not remaining_path:  # If * is the last part, return all values
                        return list(items)
                elif isinstance(current, list):
                    items = current  # type: ignore[assignment]
                    if not remaining_path:  # If * is the last part, return all items
                        return items
                else:  # Not a collection
                    raise AttributeError(
                        f"Cannot use wildcard on non-collection type: {type(current)}"
                    )

                # Apply remaining path to each item in the collection
                results = []
                for item in items:
                    try:
                        # Recursively get values from each item
                        value = self._get_value_by_path(item, remaining_path)  # type: ignore[no-untyped-call]
                        if isinstance(value, list):  # Flatten nested lists
                            results.extend(value)
                        else:
                            results.append(value)
                    except (AttributeError, KeyError):
                        # Skip items that don't have the attribute
                        pass
                return results

            # Regular attribute/key access
            elif isinstance(current, dict):
                current = current[part]
            else:
                current = getattr(current, part)

        return current



# Copyright 2025-2026 Dimensional Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import logging

from dimos.core import DimosCluster
from dimos.robot import foxglove_bridge
from dimos.robot.unitree.connection import go2
from dimos.utils.logging_config import setup_logger

logger = setup_logger(level=logging.INFO)


def deploy(dimos: DimosCluster, ip: str):  # type: ignore[no-untyped-def]
    connection = go2.deploy(dimos, ip)
    foxglove_bridge.deploy(dimos)

    # detector = moduleDB.deploy(
    #     dimos,
    #     camera=connection,
    #     lidar=connection,
    # )

    # agent = agents.deploy(dimos)
    # agent.register_skills(detector)
    return connection



# Copyright 2025-2026 Dimensional Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from __future__ import annotations

import time
from typing import TYPE_CHECKING

from pydantic import Field

if TYPE_CHECKING:
    from dimos.robot.robot import MockRobot, Robot  # type: ignore[attr-defined]
else:
    Robot = "Robot"
    MockRobot = "MockRobot"

from unitree_webrtc_connect.constants import RTC_TOPIC

from dimos.msgs.geometry_msgs import Twist, Vector3
from dimos.skills.skills import AbstractRobotSkill, AbstractSkill, SkillLibrary
from dimos.types.constants import Colors

# Module-level constant for Unitree Go2 WebRTC control definitions
UNITREE_WEBRTC_CONTROLS: list[tuple[str, int, str]] = [
    ("Damp", 1001, "Lowers the robot to the ground fully."),
    (
        "BalanceStand",
        1002,
        "Activates a mode that maintains the robot in a balanced standing position.",
    ),
    (
        "StandUp",
        1004,
        "Commands the robot to transition from a sitting or prone position to a standing posture.",
    ),
    (
        "StandDown",
        1005,
        "Instructs the robot to move from a standing position to a sitting or prone posture.",
    ),
    (
        "RecoveryStand",
        1006,
        "Recovers the robot to a state from which it can take more commands. Useful to run after multiple dynamic commands like front flips, Must run after skills like sit and jump and standup.",
    ),
    ("Sit", 1009, "Commands the robot to sit down from a standing or moving stance."),
    (
        "RiseSit",
        1010,
        "Commands the robot to rise back to a standing position from a sitting posture.",
    ),
    (
        "SwitchGait",
        1011,
        "Switches the robot's walking pattern or style dynamically, suitable for different terrains or speeds.",
    ),
    ("Trigger", 1012, "Triggers a specific action or custom routine programmed into the robot."),
    (
        "BodyHeight",
        1013,
        "Adjusts the height of the robot's body from the ground, useful for navigating various obstacles.",
    ),
    (
        "FootRaiseHeight",
        1014,
        "Controls how high the robot lifts its feet during movement, which can be adjusted for different surfaces.",
    ),
    (
        "SpeedLevel",
        1015,
        "Sets or adjusts the speed at which the robot moves, with various levels available for different operational needs.",
    ),
    (
        "Hello",
        1016,
        "Performs a greeting action, which could involve a wave or other friendly gesture.",
    ),
    ("Stretch", 1017, "Engages the robot in a stretching routine."),
    (
        "TrajectoryFollow",
        1018,
        "Directs the robot to follow a predefined trajectory, which could involve complex paths or maneuvers.",
    ),
    (
        "ContinuousGait",
        1019,
        "Enables a mode for continuous walking or running, ideal for long-distance travel.",
    ),
    ("Content", 1020, "To display or trigger when the robot is happy."),
    ("Wallow", 1021, "The robot falls onto its back and rolls around."),
    (
        "Dance1",
        1022,
        "Performs a predefined dance routine 1, programmed for entertainment or demonstration.",
    ),
    ("Dance2", 1023, "Performs another variant of a predefined dance routine 2."),
    ("GetBodyHeight", 1024, "Retrieves the current height of the robot's body from the ground."),
    (
        "GetFootRaiseHeight",
        1025,
        "Retrieves the current height at which the robot's feet are being raised during movement.",
    ),
    (
        "GetSpeedLevel",
        1026,
        "Retrieves the current speed level setting of the robot.",
    ),
    (
        "SwitchJoystick",
        1027,
        "Switches the robot's control mode to respond to joystick input for manual operation.",
    ),
    (
        "Pose",
        1028,
        "Commands the robot to assume a specific pose or posture as predefined in its programming.",
    ),
    ("Scrape", 1029, "The robot performs a scraping motion."),
    (
        "FrontFlip",
        1030,
        "Commands the robot to perform a front flip, showcasing its agility and dynamic movement capabilities.",
    ),
    (
        "FrontJump",
        1031,
        "Instructs the robot to jump forward, demonstrating its explosive movement capabilities.",
    ),
    (
        "FrontPounce",
        1032,
        "Commands the robot to perform a pouncing motion forward.",
    ),
    (
        "WiggleHips",
        1033,
        "The robot performs a hip wiggling motion, often used for entertainment or demonstration purposes.",
    ),
    (
        "GetState",
        1034,
        "Retrieves the current operational state of the robot, including its mode, position, and status.",
    ),
    (
        "EconomicGait",
        1035,
        "Engages a more energy-efficient walking or running mode to conserve battery life.",
    ),
    ("FingerHeart", 1036, "Performs a finger heart gesture while on its hind legs."),
    (
        "Handstand",
        1301,
        "Commands the robot to perform a handstand, demonstrating balance and control.",
    ),
    (
        "CrossStep",
        1302,
        "Commands the robot to perform cross-step movements.",
    ),
    (
        "OnesidedStep",
        1303,
        "Commands the robot to perform one-sided step movements.",
    ),
    ("Bound", 1304, "Commands the robot to perform bounding movements."),
    ("MoonWalk", 1305, "Commands the robot to perform a moonwalk motion."),
    ("LeftFlip", 1042, "Executes a flip towards the left side."),
    ("RightFlip", 1043, "Performs a flip towards the right side."),
    ("Backflip", 1044, "Executes a backflip, a complex and dynamic maneuver."),
]

# Module-level constants for Unitree G1 WebRTC control definitions
# G1 Arm Actions - all use api_id 7106 on topic "rt/api/arm/request"
G1_ARM_CONTROLS: list[tuple[str, int, str]] = [
    ("Handshake", 27, "Perform a handshake gesture with the right hand."),
    ("HighFive", 18, "Give a high five with the right hand."),
    ("Hug", 19, "Perform a hugging gesture with both arms."),
    ("HighWave", 26, "Wave with the hand raised high."),
    ("Clap", 17, "Clap hands together."),
    ("FaceWave", 25, "Wave near the face level."),
    ("LeftKiss", 12, "Blow a kiss with the left hand."),
    ("ArmHeart", 20, "Make a heart shape with both arms overhead."),
    ("RightHeart", 21, "Make a heart gesture with the right hand."),
    ("HandsUp", 15, "Raise both hands up in the air."),
    ("XRay", 24, "Hold arms in an X-ray pose position."),
    ("RightHandUp", 23, "Raise only the right hand up."),
    ("Reject", 22, "Make a rejection or 'no' gesture."),
    ("CancelAction", 99, "Cancel any current arm action and return hands to neutral position."),
]

# G1 Movement Modes - all use api_id 7101 on topic "rt/api/sport/request"
G1_MODE_CONTROLS: list[tuple[str, int, str]] = [
    ("WalkMode", 500, "Switch to normal walking mode."),
    ("WalkControlWaist", 501, "Switch to walking mode with waist control."),
    ("RunMode", 801, "Switch to running mode."),
]

# region MyUnitreeSkills


class MyUnitreeSkills(SkillLibrary):
    """My Unitree Skills for WebRTC interface."""

    def __init__(self, robot: Robot | None = None, robot_type: str = "go2") -> None:
        """Initialize Unitree skills library.

        Args:
            robot: Optional robot instance
            robot_type: Type of robot ("go2" or "g1"), defaults to "go2"
        """
        super().__init__()
        self._robot: Robot = None  # type: ignore[assignment]
        self.robot_type = robot_type.lower()

        if self.robot_type not in ["go2", "g1"]:
            raise ValueError(f"Unsupported robot type: {robot_type}. Must be 'go2' or 'g1'")

        # Add dynamic skills to this class based on robot type
        dynamic_skills = self.create_skills_live()
        self.register_skills(dynamic_skills)  # type: ignore[arg-type]

    @classmethod
    def register_skills(cls, skill_classes: AbstractSkill | list[AbstractSkill]) -> None:
        """Add multiple skill classes as class attributes.

        Args:
            skill_classes: List of skill classes to add
        """
        if not isinstance(skill_classes, list):
            skill_classes = [skill_classes]

        for skill_class in skill_classes:
            # Add to the class as a skill
            setattr(cls, skill_class.__name__, skill_class)  # type: ignore[attr-defined]

    def initialize_skills(self) -> None:
        for skill_class in self.get_class_skills():
            self.create_instance(skill_class.__name__, robot=self._robot)  # type: ignore[attr-defined]

        # Refresh the class skills
        self.refresh_class_skills()

    def create_skills_live(self) -> list[AbstractRobotSkill]:
        # ================================================
        # Procedurally created skills
        # ================================================
        class BaseUnitreeSkill(AbstractRobotSkill):
            """Base skill for dynamic skill creation."""

            def __call__(self) -> str:
                super().__call__()  # type: ignore[no-untyped-call]

                # For Go2: Simple api_id based call
                if hasattr(self, "_app_id"):
                    string = f"{Colors.GREEN_PRINT_COLOR}Executing Go2 skill: {self.__class__.__name__} with api_id={self._app_id}{Colors.RESET_COLOR}"
                    print(string)
                    self._robot.connection.publish_request(  # type: ignore[attr-defined]
                        RTC_TOPIC["SPORT_MOD"], {"api_id": self._app_id}
                    )
                    return f"{self.__class__.__name__} executed successfully"

                # For G1: Fixed api_id with parameter data
                elif hasattr(self, "_data_value"):
                    string = f"{Colors.GREEN_PRINT_COLOR}Executing G1 skill: {self.__class__.__name__} with data={self._data_value}{Colors.RESET_COLOR}"
                    print(string)
                    self._robot.connection.publish_request(  # type: ignore[attr-defined]
                        self._topic,  # type: ignore[attr-defined]
                        {"api_id": self._api_id, "parameter": {"data": self._data_value}},  # type: ignore[attr-defined]
                    )
                    return f"{self.__class__.__name__} executed successfully"
                else:
                    raise RuntimeError(
                        f"Skill {self.__class__.__name__} missing required attributes"
                    )

        skills_classes = []

        if self.robot_type == "g1":
            # Create G1 arm skills
            for name, data_value, description in G1_ARM_CONTROLS:
                skill_class = type(
                    name,
                    (BaseUnitreeSkill,),
                    {
                        "__doc__": description,
                        "_topic": "rt/api/arm/request",
                        "_api_id": 7106,
                        "_data_value": data_value,
                    },
                )
                skills_classes.append(skill_class)

            # Create G1 mode skills
            for name, data_value, description in G1_MODE_CONTROLS:
                skill_class = type(
                    name,
                    (BaseUnitreeSkill,),
                    {
                        "__doc__": description,
                        "_topic": "rt/api/sport/request",
                        "_api_id": 7101,
                        "_data_value": data_value,
                    },
                )
                skills_classes.append(skill_class)
        else:
            # Go2 skills (existing code)
            for name, app_id, description in UNITREE_WEBRTC_CONTROLS:
                if name not in ["Reverse", "Spin"]:  # Exclude reverse and spin skills
                    skill_class = type(
                        name, (BaseUnitreeSkill,), {"__doc__": description, "_app_id": app_id}
                    )
                    skills_classes.append(skill_class)

        return skills_classes  # type: ignore[return-value]

    # region Class-based Skills

    class Move(AbstractRobotSkill):
        """Move the robot using direct velocity commands. Determine duration required based on user distance instructions."""

        x: float = Field(..., description="Forward velocity (m/s).")
        y: float = Field(default=0.0, description="Left/right velocity (m/s)")
        yaw: float = Field(default=0.0, description="Rotational velocity (rad/s)")
        duration: float = Field(default=0.0, description="How long to move (seconds).")

        def __call__(self) -> str:
            self._robot.move(  # type: ignore[attr-defined]
                Twist(linear=Vector3(self.x, self.y, 0.0), angular=Vector3(0.0, 0.0, self.yaw)),
                duration=self.duration,
            )
            return f"started moving with velocity={self.x}, {self.y}, {self.yaw} for {self.duration} seconds"

    class Wait(AbstractSkill):
        """Wait for a specified amount of time."""

        seconds: float = Field(..., description="Seconds to wait")

        def __call__(self) -> str:
            time.sleep(self.seconds)
            return f"Wait completed with length={self.seconds}s"

    # endregion


# endregion



# Copyright 2025-2026 Dimensional Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""Mock backend for testing manipulator drivers without hardware.

Usage:
    >>> from dimos.hardware.manipulators.xarm import XArm
    >>> from dimos.hardware.manipulators.mock import MockBackend
    >>> arm = XArm(backend=MockBackend())
    >>> arm.start()  # No hardware needed!
    >>> arm.move_joint([0.1, 0.2, 0.3, 0.4, 0.5, 0.6])
    >>> assert arm.backend.read_joint_positions() == [0.1, 0.2, 0.3, 0.4, 0.5, 0.6]
"""

from dimos.hardware.manipulators.mock.backend import MockBackend

__all__ = ["MockBackend"]



# Copyright 2025-2026 Dimensional Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from collections.abc import Callable
from functools import wraps
import threading
import time
from typing import Any, Protocol, TypeVar

from .accumulators import Accumulator, LatestAccumulator

_CacheResult_co = TypeVar("_CacheResult_co", covariant=True)
_CacheReturn = TypeVar("_CacheReturn")


class CachedMethod(Protocol[_CacheResult_co]):
    """Protocol for methods decorated with simple_mcache."""

    def __call__(self) -> _CacheResult_co: ...
    def invalidate_cache(self, instance: Any) -> None: ...


def limit(max_freq: float, accumulator: Accumulator | None = None):  # type: ignore[no-untyped-def, type-arg]
    """
    Decorator that limits function call frequency.

    If calls come faster than max_freq, they are skipped.
    If calls come slower than max_freq, they pass through immediately.

    Args:
        max_freq: Maximum frequency in Hz (calls per second)
        accumulator: Optional accumulator to collect skipped calls (defaults to LatestAccumulator)

    Returns:
        Decorated function that respects the frequency limit
    """
    if max_freq <= 0:
        raise ValueError("Frequency must be positive")

    min_interval = 1.0 / max_freq

    # Create default accumulator if none provided
    if accumulator is None:
        accumulator = LatestAccumulator()

    def decorator(func: Callable) -> Callable:  # type: ignore[type-arg]
        last_call_time = 0.0
        lock = threading.Lock()
        timer: threading.Timer | None = None

        def execute_accumulated() -> None:
            nonlocal last_call_time, timer
            with lock:
                if len(accumulator):
                    acc_args, acc_kwargs = accumulator.get()  # type: ignore[misc]
                    last_call_time = time.time()
                    timer = None
                    func(*acc_args, **acc_kwargs)

        @wraps(func)
        def wrapper(*args, **kwargs):  # type: ignore[no-untyped-def]
            nonlocal last_call_time, timer
            current_time = time.time()

            with lock:
                time_since_last = current_time - last_call_time

                if time_since_last >= min_interval:
                    # Cancel any pending timer
                    if timer is not None:
                        timer.cancel()
                        timer = None

                    # Enough time has passed, execute the function
                    last_call_time = current_time

                    # if we have accumulated data, we get a compound value
                    if len(accumulator):
                        accumulator.add(*args, **kwargs)
                        acc_args, acc_kwargs = accumulator.get()  # type: ignore[misc]  # accumulator resets here
                        return func(*acc_args, **acc_kwargs)

                    # No accumulated data, normal call
                    return func(*args, **kwargs)

                else:
                    # Too soon, skip this call
                    accumulator.add(*args, **kwargs)

                    # Schedule execution for when the interval expires
                    if timer is not None:
                        timer.cancel()

                    time_to_wait = min_interval - time_since_last
                    timer = threading.Timer(time_to_wait, execute_accumulated)
                    timer.start()

                    return None

        return wrapper

    return decorator


def simple_mcache(method: Callable) -> Callable:  # type: ignore[type-arg]
    """
    Decorator to cache the result of a method call on the instance.

    The cached value is stored as an attribute on the instance with the name
    `_cached_<method_name>`. Subsequent calls to the method will return the
    cached value instead of recomputing it.

    Thread-safe: Uses a lock per instance to ensure the cached value is
    computed only once even in multi-threaded environments.

    Args:
        method: The method to be decorated.

    Returns:
        The decorated method with caching behavior.
    """

    attr_name = f"_cached_{method.__name__}"
    lock_name = f"_lock_{method.__name__}"

    @wraps(method)
    def getter(self):  # type: ignore[no-untyped-def]
        # Get or create the lock for this instance
        if not hasattr(self, lock_name):
            setattr(self, lock_name, threading.Lock())

        lock = getattr(self, lock_name)

        if hasattr(self, attr_name):
            return getattr(self, attr_name)

        with lock:
            # Check again inside the lock
            if not hasattr(self, attr_name):
                setattr(self, attr_name, method(self))
            return getattr(self, attr_name)

    def invalidate_cache(instance: Any) -> None:
        """Clear the cached value for the given instance."""
        if not hasattr(instance, lock_name):
            return

        lock = getattr(instance, lock_name)
        with lock:
            if hasattr(instance, attr_name):
                delattr(instance, attr_name)

    getter.invalidate_cache = invalidate_cache  # type: ignore[attr-defined]

    return getter


def retry(max_retries: int = 3, on_exception: type[Exception] = Exception, delay: float = 0.0):  # type: ignore[no-untyped-def]
    """
    Decorator that retries a function call if it raises an exception.

    Args:
        max_retries: Maximum number of retry attempts (default: 3)
        on_exception: Exception type to catch and retry on (default: Exception)
        delay: Fixed delay in seconds between retries (default: 0.0)

    Returns:
        Decorated function that will retry on failure

    Example:
        @retry(max_retries=5, on_exception=ConnectionError, delay=0.5)
        def connect_to_server():
            # connection logic that might fail
            pass

        @retry()  # Use defaults: 3 retries on any Exception, no delay
        def risky_operation():
            # might fail occasionally
            pass
    """
    if max_retries < 0:
        raise ValueError("max_retries must be non-negative")
    if delay < 0:
        raise ValueError("delay must be non-negative")

    def decorator(func: Callable) -> Callable:  # type: ignore[type-arg]
        @wraps(func)
        def wrapper(*args, **kwargs):  # type: ignore[no-untyped-def]
            last_exception = None

            for attempt in range(max_retries + 1):
                try:
                    return func(*args, **kwargs)
                except on_exception as e:
                    last_exception = e
                    if attempt < max_retries:
                        # Still have retries left
                        if delay > 0:
                            time.sleep(delay)
                        continue
                    else:
                        # Out of retries, re-raise the last exception
                        raise

            # This should never be reached, but just in case
            if last_exception:
                raise last_exception

        return wrapper

    return decorator



# Copyright 2025-2026 Dimensional Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from typing import Protocol

from dimos.core import In, Out
from dimos.msgs.geometry_msgs import PoseStamped, Twist
from dimos.msgs.nav_msgs import Path


class Nav(Protocol):
    goal_req: In[PoseStamped]
    goal_active: Out[PoseStamped]
    path_active: Out[Path]
    ctrl: Out[Twist]

    # identity quaternion (Quaternion(0,0,0,1)) represents "no rotation requested"
    def navigate_to_target(self, target: PoseStamped) -> None: ...

    def stop_navigating(self) -> None: ...



# Copyright 2025-2026 Dimensional Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from threading import RLock
import time
from typing import cast

import numpy as np
from numpy.typing import NDArray

from dimos.msgs.geometry_msgs.PoseStamped import PoseStamped

_max_points_per_second = 1000


class PositionTracker:
    _lock: RLock
    _time_window: float
    _max_points: int
    _threshold: float
    _timestamps: NDArray[np.float32]
    _positions: NDArray[np.float32]
    _index: int
    _size: int

    def __init__(self, time_window: float) -> None:
        self._lock = RLock()
        self._time_window = time_window
        self._threshold = 0.4
        self._max_points = int(_max_points_per_second * self._time_window)
        self.reset_data()

    def reset_data(self) -> None:
        with self._lock:
            self._timestamps = np.zeros(self._max_points, dtype=np.float32)
            self._positions = np.zeros((self._max_points, 2), dtype=np.float32)
            self._index = 0
            self._size = 0

    def add_position(self, pose: PoseStamped) -> None:
        with self._lock:
            self._timestamps[self._index] = time.time()
            self._positions[self._index] = (pose.position.x, pose.position.y)
            self._index = (self._index + 1) % self._max_points
            self._size = min(self._size + 1, self._max_points)

    def _get_recent_positions(self) -> NDArray[np.float32]:
        cutoff = time.time() - self._time_window

        if self._size == 0:
            return np.empty((0, 2), dtype=np.float32)

        if self._size < self._max_points:
            mask = self._timestamps[: self._size] >= cutoff
            return self._positions[: self._size][mask]

        ts = np.concatenate([self._timestamps[self._index :], self._timestamps[: self._index]])
        pos = np.concatenate([self._positions[self._index :], self._positions[: self._index]])
        mask = ts >= cutoff
        return cast("NDArray[np.float32]", pos[mask])

    def is_stuck(self) -> bool:
        with self._lock:
            recent = self._get_recent_positions()

        if len(recent) == 0:
            return False

        centroid = recent.mean(axis=0)
        distances = np.linalg.norm(recent - centroid, axis=1)

        return bool(np.all(distances < self._threshold))



# Copyright 2025-2026 Dimensional Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from pathlib import Path

DIMOS_PROJECT_ROOT = Path(__file__).parent.parent

DIMOS_LOG_DIR = DIMOS_PROJECT_ROOT / "logs"

"""
Constants for shared memory
Usually, auto-detection for size would be preferred. Sadly, though, channels are made
and frozen *before* the first frame is received.
Therefore, a maximum capacity for color image and depth image transfer should be defined
ahead of time.
"""
# Default color image size: 1920x1080 frame x 3 (RGB) x uint8
DEFAULT_CAPACITY_COLOR_IMAGE = 1920 * 1080 * 3
# Default depth image size: 1280x720 frame * 4 (float32 size)
DEFAULT_CAPACITY_DEPTH_IMAGE = 1280 * 720 * 4

# From https://github.com/lcm-proj/lcm.git
LCM_MAX_CHANNEL_NAME_LENGTH = 63



# Copyright 2025-2026 Dimensional Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from dimos.stream.audio.node_key_recorder import KeyRecorder
from dimos.stream.audio.node_microphone import SounddeviceAudioSource
from dimos.stream.audio.node_normalizer import AudioNormalizer
from dimos.stream.audio.node_output import SounddeviceAudioOutput
from dimos.stream.audio.node_volume_monitor import monitor
from dimos.stream.audio.stt.node_whisper import WhisperNode
from dimos.stream.audio.text.node_stdout import TextPrinterNode
from dimos.stream.audio.tts.node_openai import OpenAITTSNode, Voice


def stt():  # type: ignore[no-untyped-def]
    # Create microphone source, recorder, and audio output
    mic = SounddeviceAudioSource()
    normalizer = AudioNormalizer()
    recorder = KeyRecorder(always_subscribe=True)
    whisper_node = WhisperNode()  # Assign to global variable

    # Connect audio processing pipeline
    normalizer.consume_audio(mic.emit_audio())
    recorder.consume_audio(normalizer.emit_audio())
    monitor(recorder.emit_audio())
    whisper_node.consume_audio(recorder.emit_recording())

    user_text_printer = TextPrinterNode(prefix="USER: ")
    user_text_printer.consume_text(whisper_node.emit_text())

    return whisper_node


def tts():  # type: ignore[no-untyped-def]
    tts_node = OpenAITTSNode(speed=1.2, voice=Voice.ONYX)
    agent_text_printer = TextPrinterNode(prefix="AGENT: ")
    agent_text_printer.consume_text(tts_node.emit_text())

    response_output = SounddeviceAudioOutput(sample_rate=24000)
    response_output.consume_audio(tts_node.emit_audio())

    return tts_node



# Copyright 2025-2026 Dimensional Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import hashlib

from rich.console import Console
from rich.table import Table
from rich.text import Text

from dimos.types.timestamped import to_timestamp


def _hash_to_color(name: str) -> str:
    """Generate a consistent color for a given name using hash."""
    # List of rich colors to choose from
    colors = [
        "cyan",
        "magenta",
        "yellow",
        "blue",
        "green",
        "red",
        "bright_cyan",
        "bright_magenta",
        "bright_yellow",
        "bright_blue",
        "bright_green",
        "bright_red",
        "purple",
        "white",
        "pink",
    ]

    # Hash the name and pick a color
    hash_value = hashlib.md5(name.encode()).digest()[0]
    return colors[hash_value % len(colors)]


class TableStr:
    """Mixin class that provides table-based string representation for detection collections."""

    def __str__(self) -> str:
        console = Console(force_terminal=True, legacy_windows=False)

        # Create a table for detections
        table = Table(
            title=f"{self.__class__.__name__} [{len(self.detections)} detections @ {to_timestamp(self.image.ts):.3f}]",  # type: ignore[attr-defined]
            show_header=True,
            show_edge=True,
        )

        # Dynamically build columns based on the first detection's dict keys
        if not self.detections:  # type: ignore[attr-defined]
            return (
                f"   {self.__class__.__name__} [0 detections @ {to_timestamp(self.image.ts):.3f}]"  # type: ignore[attr-defined]
            )

        # Cache all repr_dicts to avoid double computation
        detection_dicts = [det.to_repr_dict() for det in self]  # type: ignore[attr-defined]

        first_dict = detection_dicts[0]
        table.add_column("#", style="dim")
        for col in first_dict.keys():
            color = _hash_to_color(col)
            table.add_column(col.title(), style=color)

        # Add each detection to the table
        for i, d in enumerate(detection_dicts):
            row = [str(i)]

            for key in first_dict.keys():
                if key == "conf":
                    # Color-code confidence
                    conf_color = (
                        "green"
                        if float(d[key]) > 0.8
                        else "yellow"
                        if float(d[key]) > 0.5
                        else "red"
                    )
                    row.append(Text(f"{d[key]}", style=conf_color))  # type: ignore[arg-type]
                elif key == "points" and d.get(key) == "None":
                    row.append(Text(d.get(key, ""), style="dim"))  # type: ignore[arg-type]
                else:
                    row.append(str(d.get(key, "")))
            table.add_row(*row)

        with console.capture() as capture:
            console.print(table)
        return capture.get().strip()



#!/usr/bin/env python3
# Copyright 2025-2026 Dimensional Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from enum import Enum
import io
import threading
import time

from openai import OpenAI
from reactivex import Observable, Subject
import soundfile as sf  # type: ignore[import-untyped]

from dimos.stream.audio.base import (
    AbstractAudioEmitter,
    AudioEvent,
)
from dimos.stream.audio.text.base import AbstractTextConsumer, AbstractTextEmitter
from dimos.utils.logging_config import setup_logger

logger = setup_logger()


class Voice(str, Enum):
    """Available voices in OpenAI TTS API."""

    ALLOY = "alloy"
    ECHO = "echo"
    FABLE = "fable"
    ONYX = "onyx"
    NOVA = "nova"
    SHIMMER = "shimmer"


class OpenAITTSNode(AbstractTextConsumer, AbstractAudioEmitter, AbstractTextEmitter):
    """
    A text-to-speech node that consumes text, emits audio using OpenAI's TTS API, and passes through text.

    This node implements AbstractTextConsumer to receive text input, AbstractAudioEmitter
    to provide audio output, and AbstractTextEmitter to pass through the text being spoken,
    allowing it to be inserted into a text-to-audio pipeline with text passthrough capabilities.
    """

    def __init__(
        self,
        api_key: str | None = None,
        voice: Voice = Voice.ECHO,
        model: str = "tts-1",
        buffer_size: int = 1024,
        speed: float = 1.0,
    ) -> None:
        """
        Initialize OpenAITTSNode.

        Args:
            api_key: OpenAI API key (if None, will try to use environment variable)
            voice: TTS voice to use
            model: TTS model to use
            buffer_size: Audio buffer size in samples
        """
        self.voice = voice
        self.model = model
        self.speed = speed
        self.buffer_size = buffer_size

        # Initialize OpenAI client
        self.client = OpenAI(api_key=api_key)

        # Initialize state
        self.audio_subject = Subject()  # type: ignore[var-annotated]
        self.text_subject = Subject()  # type: ignore[var-annotated]
        self.subscription = None
        self.processing_thread = None
        self.is_running = True
        self.text_queue = []  # type: ignore[var-annotated]
        self.queue_lock = threading.Lock()

    def emit_audio(self) -> Observable:  # type: ignore[type-arg]
        """
        Returns an observable that emits audio frames.

        Returns:
            Observable emitting AudioEvent objects
        """
        return self.audio_subject

    def emit_text(self) -> Observable:  # type: ignore[type-arg]
        """
        Returns an observable that emits the text being spoken.

        Returns:
            Observable emitting text strings
        """
        return self.text_subject

    def consume_text(self, text_observable: Observable) -> "AbstractTextConsumer":  # type: ignore[type-arg]
        """
        Start consuming text from the observable source.

        Args:
            text_observable: Observable source of text strings

        Returns:
            Self for method chaining
        """
        logger.info("Starting OpenAITTSNode")

        # Start the processing thread
        self.processing_thread = threading.Thread(target=self._process_queue, daemon=True)  # type: ignore[assignment]
        self.processing_thread.start()  # type: ignore[attr-defined]

        # Subscribe to the text observable
        self.subscription = text_observable.subscribe(  # type: ignore[assignment]
            on_next=self._queue_text,
            on_error=lambda e: logger.error(f"Error in OpenAITTSNode: {e}"),
        )

        return self

    def _queue_text(self, text: str) -> None:
        """
        Add text to the processing queue and pass it through to text_subject.

        Args:
            text: The text to synthesize
        """
        if not text.strip():
            return

        with self.queue_lock:
            self.text_queue.append(text)

    def _process_queue(self) -> None:
        """Background thread to process the text queue."""
        while self.is_running:
            # Check if there's text to process
            text_to_process = None
            with self.queue_lock:
                if self.text_queue:
                    text_to_process = self.text_queue.pop(0)

            if text_to_process:
                self._synthesize_speech(text_to_process)
            else:
                # Sleep a bit to avoid busy-waiting
                time.sleep(0.1)

    def _synthesize_speech(self, text: str) -> None:
        """
        Convert text to speech using OpenAI API.

        Args:
            text: The text to synthesize
        """
        try:
            # Call OpenAI TTS API
            response = self.client.audio.speech.create(
                model=self.model, voice=self.voice.value, input=text, speed=self.speed
            )
            self.text_subject.on_next(text)

            # Convert the response to audio data
            audio_data = io.BytesIO(response.content)

            # Read with soundfile
            with sf.SoundFile(audio_data, "r") as sound_file:
                # Get the sample rate from the file
                actual_sample_rate = sound_file.samplerate
                # Read the entire file
                audio_array = sound_file.read()

            # Debug log the sample rate from the OpenAI file
            logger.debug(f"OpenAI audio sample rate: {actual_sample_rate}Hz")

            timestamp = time.time()

            # Create AudioEvent and emit it
            audio_event = AudioEvent(
                data=audio_array,
                sample_rate=24000,
                timestamp=timestamp,
                channels=1 if audio_array.ndim == 1 else audio_array.shape[1],
            )

            self.audio_subject.on_next(audio_event)

        except Exception as e:
            logger.error(f"Error synthesizing speech: {e}")

    def dispose(self) -> None:
        """Clean up resources."""
        logger.info("Disposing OpenAITTSNode")

        self.is_running = False

        if self.processing_thread and self.processing_thread.is_alive():
            self.processing_thread.join(timeout=5.0)

        if self.subscription:
            self.subscription.dispose()
            self.subscription = None

        # Complete the subjects
        self.audio_subject.on_completed()
        self.text_subject.on_completed()


if __name__ == "__main__":
    import time

    from reactivex import Subject

    from dimos.stream.audio.node_output import SounddeviceAudioOutput
    from dimos.stream.audio.text.node_stdout import TextPrinterNode
    from dimos.stream.audio.utils import keepalive

    # Create a simple text subject that we can push values to
    text_subject = Subject()  # type: ignore[var-annotated]

    tts_node = OpenAITTSNode(voice=Voice.ALLOY)
    tts_node.consume_text(text_subject)

    # Create and connect an audio output node - explicitly set sample rate
    audio_output = SounddeviceAudioOutput(sample_rate=24000)
    audio_output.consume_audio(tts_node.emit_audio())

    stdout = TextPrinterNode(prefix="[Spoken Text] ")

    stdout.consume_text(tts_node.emit_text())

    # Emit some test messages
    test_messages = [
        "Hello!",
        "This is a test of the OpenAI text to speech system.",
    ]

    print("Starting OpenAI TTS test...")
    print("-" * 60)

    for _i, message in enumerate(test_messages):
        text_subject.on_next(message)

    keepalive()



# Copyright 2025-2026 Dimensional Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from __future__ import annotations

from rich.text import Text
from textual.app import App, ComposeResult
from textual.color import Color
from textual.widgets import DataTable

from dimos.utils.cli import theme
from dimos.utils.cli.lcmspy.lcmspy import GraphLCMSpy, GraphTopic as SpyTopic


def gradient(max_value: float, value: float) -> str:
    """Gradient from cyan (low) to yellow (high) using DimOS theme colors"""
    ratio = min(value / max_value, 1.0)
    # Parse hex colors from theme
    cyan = Color.parse(theme.CYAN)
    yellow = Color.parse(theme.YELLOW)
    color = cyan.blend(yellow, ratio)

    return color.hex


def topic_text(topic_name: str) -> Text:
    """Format topic name with DimOS theme colors"""
    if "#" in topic_name:
        parts = topic_name.split("#", 1)
        return Text(parts[0], style=theme.BRIGHT_WHITE) + Text("#" + parts[1], style=theme.BLUE)

    if topic_name[:4] == "/rpc":
        return Text(topic_name[:4], style=theme.BLUE) + Text(
            topic_name[4:], style=theme.BRIGHT_WHITE
        )

    return Text(topic_name, style=theme.BRIGHT_WHITE)


class LCMSpyApp(App):  # type: ignore[type-arg]
    """A real-time CLI dashboard for LCM traffic statistics using Textual."""

    CSS_PATH = "../dimos.tcss"

    CSS = f"""
    Screen {{
        layout: vertical;
        background: {theme.BACKGROUND};
    }}
    DataTable {{
        height: 2fr;
        width: 1fr;
        border: solid {theme.BORDER};
        background: {theme.BG};
        scrollbar-size: 0 0;
    }}
    DataTable > .datatable--header {{
        color: {theme.ACCENT};
        background: transparent;
    }}
    """

    refresh_interval: float = 0.5  # seconds

    BINDINGS = [
        ("q", "quit"),
        ("ctrl+c", "quit"),
    ]

    def __init__(self, *args, **kwargs) -> None:  # type: ignore[no-untyped-def]
        super().__init__(*args, **kwargs)
        self.spy = GraphLCMSpy(autoconf=True, graph_log_window=0.5)
        self.table: DataTable | None = None  # type: ignore[type-arg]

    def compose(self) -> ComposeResult:
        self.table = DataTable(zebra_stripes=False, cursor_type=None)  # type: ignore[arg-type]
        self.table.add_column("Topic")
        self.table.add_column("Freq (Hz)")
        self.table.add_column("Bandwidth")
        self.table.add_column("Total Traffic")
        yield self.table

    def on_mount(self) -> None:
        self.spy.start()
        self.set_interval(self.refresh_interval, self.refresh_table)

    async def on_unmount(self) -> None:
        self.spy.stop()

    def refresh_table(self) -> None:
        topics: list[SpyTopic] = list(self.spy.topic.values())  # type: ignore[arg-type, call-arg]
        topics.sort(key=lambda t: t.total_traffic(), reverse=True)
        self.table.clear(columns=False)  # type: ignore[union-attr]

        for t in topics:
            freq = t.freq(5.0)
            kbps = t.kbps(5.0)
            bw_val, bw_unit = t.kbps_hr(5.0)
            total_val, total_unit = t.total_traffic_hr()

            self.table.add_row(  # type: ignore[union-attr]
                topic_text(t.name),
                Text(f"{freq:.1f}", style=gradient(10, freq)),
                Text(f"{bw_val} {bw_unit.value}/s", style=gradient(1024 * 3, kbps)),
                Text(f"{total_val} {total_unit.value}"),
            )


def main() -> None:
    import sys

    if len(sys.argv) > 1 and sys.argv[1] == "web":
        import os

        from textual_serve.server import Server  # type: ignore[import-not-found]

        server = Server(f"python {os.path.abspath(__file__)}")
        server.serve()
    else:
        LCMSpyApp().run()


if __name__ == "__main__":
    main()



#!/usr/bin/env python3
# Copyright 2025-2026 Dimensional Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from dotenv import load_dotenv

from dimos.agents.agent import llm_agent
from dimos.agents.cli.human import human_input
from dimos.agents.skills.demo_calculator_skill import demo_calculator_skill
from dimos.core.blueprints import autoconnect

load_dotenv()


demo_skill = autoconnect(
    demo_calculator_skill(),
    human_input(),
    llm_agent(),
)



#!/usr/bin/env python3
# Copyright 2025-2026 Dimensional Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

# Copyright 2025-2026 Dimensional Inc.

"""Int32 message type."""

from typing import ClassVar

from dimos_lcm.std_msgs import Int32 as LCMInt32


class Int32(LCMInt32):  # type: ignore[misc]
    """ROS-compatible Int32 message."""

    msg_name: ClassVar[str] = "std_msgs.Int32"

    def __init__(self, data: int = 0) -> None:
        """Initialize Int32 with data value."""
        self.data = data



# Copyright 2025-2026 Dimensional Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""
Joint Trajectory Controller

A simple joint-space trajectory executor. Does NOT:
- Use Cartesian space
- Compute error
- Apply PID
- Call IK

Just samples a trajectory at time t and sends joint positions to the driver.

Behavior:
- execute_trajectory(): Preempts any active trajectory, starts new one immediately
- cancel(): Stops at current position
- reset(): Required to recover from FAULT state
"""

from dataclasses import dataclass
import threading
import time
from typing import Any

from dimos.core import In, Module, Out, rpc
from dimos.core.module import ModuleConfig
from dimos.msgs.sensor_msgs import JointCommand, JointState, RobotState
from dimos.msgs.trajectory_msgs import JointTrajectory, TrajectoryState, TrajectoryStatus
from dimos.utils.logging_config import setup_logger

logger = setup_logger()


@dataclass
class JointTrajectoryControllerConfig(ModuleConfig):
    """Configuration for joint trajectory controller."""

    control_frequency: float = 100.0  # Hz - trajectory execution rate


class JointTrajectoryController(Module):
    """
    Joint-space trajectory executor.

    Executes joint trajectories at 100Hz by sampling and forwarding
    joint positions to the arm driver. Uses ROS action-server-like
    state machine for execution control.

    State Machine:
        IDLE ──execute()──► EXECUTING ──done──► COMPLETED
          ▲                     │                    │
          │                  cancel()             reset()
          │                     ▼                    │
          └─────reset()───── ABORTED ◄──────────────┘
                                │
                             error
                                ▼
                              FAULT ──reset()──► IDLE
    """

    default_config = JointTrajectoryControllerConfig
    config: JointTrajectoryControllerConfig  # Type hint for proper attribute access

    # Input topics
    joint_state: In[JointState] = None  # type: ignore[assignment]  # Feedback from arm driver
    robot_state: In[RobotState] = None  # type: ignore[assignment]  # Robot status from arm driver
    trajectory: In[JointTrajectory] = None  # type: ignore[assignment]  # Trajectory to execute (topic-based)

    # Output topics
    joint_position_command: Out[JointCommand] = None  # type: ignore[assignment]  # To arm driver

    def __init__(self, *args: Any, **kwargs: Any) -> None:
        super().__init__(*args, **kwargs)

        # State machine
        self._state = TrajectoryState.IDLE
        self._lock = threading.Lock()

        # Active trajectory
        self._trajectory: JointTrajectory | None = None
        self._start_time: float = 0.0

        # Latest feedback
        self._latest_joint_state: JointState | None = None
        self._latest_robot_state: RobotState | None = None

        # Error tracking
        self._error_message: str = ""

        # Execution thread
        self._exec_thread: threading.Thread | None = None
        self._stop_event = threading.Event()

        logger.info(f"JointTrajectoryController initialized at {self.config.control_frequency}Hz")

    @rpc
    def start(self) -> None:
        """Start the trajectory controller."""
        super().start()

        # Subscribe to feedback topics
        try:
            if self.joint_state.connection is not None or self.joint_state._transport is not None:
                self.joint_state.subscribe(self._on_joint_state)
                logger.info("Subscribed to joint_state")
        except Exception as e:
            logger.warning(f"Failed to subscribe to joint_state: {e}")

        try:
            if self.robot_state.connection is not None or self.robot_state._transport is not None:
                self.robot_state.subscribe(self._on_robot_state)
                logger.info("Subscribed to robot_state")
        except Exception as e:
            logger.warning(f"Failed to subscribe to robot_state: {e}")

        # Subscribe to trajectory topic
        try:
            if self.trajectory.connection is not None or self.trajectory._transport is not None:
                self.trajectory.subscribe(self._on_trajectory)
                logger.info("Subscribed to trajectory topic")
        except Exception:
            logger.debug("trajectory topic not connected (expected - can use RPC instead)")

        # Start execution thread
        self._stop_event.clear()
        self._exec_thread = threading.Thread(
            target=self._execution_loop, daemon=True, name="trajectory_exec_thread"
        )
        self._exec_thread.start()

        logger.info("JointTrajectoryController started")

    @rpc
    def stop(self) -> None:
        """Stop the trajectory controller."""
        logger.info("Stopping JointTrajectoryController...")

        self._stop_event.set()

        if self._exec_thread and self._exec_thread.is_alive():
            self._exec_thread.join(timeout=2.0)

        super().stop()
        logger.info("JointTrajectoryController stopped")

    # =========================================================================
    # RPC Methods - Action-server-like interface
    # =========================================================================

    @rpc
    def execute_trajectory(self, trajectory: JointTrajectory) -> bool:
        """
        Set and start executing a new trajectory immediately.
        If currently executing, preempts and starts new trajectory.

        Args:
            trajectory: JointTrajectory to execute

        Returns:
            True if accepted, False if in FAULT state or trajectory invalid
        """
        with self._lock:
            # Cannot execute if in FAULT state
            if self._state == TrajectoryState.FAULT:
                logger.warning(
                    "Cannot execute trajectory: controller in FAULT state (call reset())"
                )
                return False

            # Validate trajectory
            if trajectory is None or trajectory.duration <= 0:
                logger.warning("Invalid trajectory: None or zero duration")
                return False

            if not trajectory.points:
                logger.warning("Invalid trajectory: no points")
                return False

            # Preempt any active trajectory
            if self._state == TrajectoryState.EXECUTING:
                logger.info("Preempting active trajectory")

            # Start new trajectory
            self._trajectory = trajectory
            self._start_time = time.time()
            self._state = TrajectoryState.EXECUTING
            self._error_message = ""

            logger.info(
                f"Executing trajectory: {len(trajectory.points)} points, "
                f"duration={trajectory.duration:.3f}s"
            )
            return True

    @rpc
    def cancel(self) -> bool:
        """
        Cancel the currently executing trajectory.
        Robot stops at current position.

        Returns:
            True if cancelled, False if no active trajectory
        """
        with self._lock:
            if self._state != TrajectoryState.EXECUTING:
                logger.debug("No active trajectory to cancel")
                return False

            self._state = TrajectoryState.ABORTED
            logger.info("Trajectory cancelled")
            return True

    @rpc
    def reset(self) -> bool:
        """
        Reset from FAULT, COMPLETED, or ABORTED state back to IDLE.
        Required before executing new trajectories after a fault.

        Returns:
            True if reset successful, False if currently EXECUTING
        """
        with self._lock:
            if self._state == TrajectoryState.EXECUTING:
                logger.warning("Cannot reset while executing (call cancel() first)")
                return False

            self._state = TrajectoryState.IDLE
            self._trajectory = None
            self._error_message = ""
            logger.info("Controller reset to IDLE")
            return True

    @rpc
    def get_status(self) -> TrajectoryStatus:
        """
        Get the current status of the trajectory execution.

        Returns:
            TrajectoryStatus with state, progress, and error info
        """
        with self._lock:
            time_elapsed = 0.0
            time_remaining = 0.0
            progress = 0.0

            if self._trajectory is not None and self._state == TrajectoryState.EXECUTING:
                time_elapsed = time.time() - self._start_time
                time_remaining = max(0.0, self._trajectory.duration - time_elapsed)
                progress = (
                    min(1.0, time_elapsed / self._trajectory.duration)
                    if self._trajectory.duration > 0
                    else 1.0
                )

            return TrajectoryStatus(
                state=self._state,
                progress=progress,
                time_elapsed=time_elapsed,
                time_remaining=time_remaining,
                error=self._error_message,
            )

    # =========================================================================
    # Callbacks
    # =========================================================================

    def _on_joint_state(self, msg: JointState) -> None:
        """Callback for joint state feedback."""
        self._latest_joint_state = msg

    def _on_robot_state(self, msg: RobotState) -> None:
        """Callback for robot state feedback."""
        self._latest_robot_state = msg

    def _on_trajectory(self, msg: JointTrajectory) -> None:
        """Callback when trajectory is received via topic."""
        logger.info(
            f"Received trajectory via topic: {len(msg.points)} points, duration={msg.duration:.3f}s"
        )
        self.execute_trajectory(msg)

    # =========================================================================
    # Execution Loop
    # =========================================================================

    def _execution_loop(self) -> None:
        """
        Main execution loop running at control_frequency Hz.

        When EXECUTING:
        1. Compute elapsed time
        2. Sample trajectory at t
        3. Publish joint command
        4. Check if done
        """
        period = 1.0 / self.config.control_frequency
        logger.info(f"Execution loop started at {self.config.control_frequency}Hz")

        while not self._stop_event.is_set():
            try:
                with self._lock:
                    # Only process if executing
                    if self._state != TrajectoryState.EXECUTING:
                        # Release lock and sleep
                        pass
                    else:
                        # Compute elapsed time
                        t = time.time() - self._start_time

                        # Check if trajectory complete
                        if self._trajectory is None:
                            self._state = TrajectoryState.FAULT
                            logger.error("Trajectory is None during execution")
                        elif t >= self._trajectory.duration:
                            self._state = TrajectoryState.COMPLETED
                            logger.info(
                                f"Trajectory completed: duration={self._trajectory.duration:.3f}s"
                            )
                        else:
                            # Sample trajectory
                            q_ref, _qd_ref = self._trajectory.sample(t)

                            # Create and publish command (outside lock would be better but simpler here)
                            cmd = JointCommand(positions=q_ref, timestamp=time.time())

                            # Publish - must release lock first for thread safety
                            trajectory_active = True

                if trajectory_active if "trajectory_active" in dir() else False:
                    try:
                        self.joint_position_command.publish(cmd)
                    except Exception as e:
                        logger.error(f"Failed to publish joint command: {e}")
                        with self._lock:
                            self._state = TrajectoryState.FAULT
                            self._error_message = f"Publish failed: {e}"

                # Reset flag
                trajectory_active = False

                # Maintain loop frequency
                time.sleep(period)

            except Exception as e:
                logger.error(f"Error in execution loop: {e}")
                with self._lock:
                    if self._state == TrajectoryState.EXECUTING:
                        self._state = TrajectoryState.FAULT
                        self._error_message = str(e)
                time.sleep(period)

        logger.info("Execution loop stopped")


# Expose blueprint for declarative composition
joint_trajectory_controller = JointTrajectoryController.blueprint



#!/usr/bin/env python3
# Copyright 2025-2026 Dimensional Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import itertools

from dimos.msgs.sensor_msgs import PointCloud2
from dimos.robot.unitree_webrtc.type.lidar import pointcloud2_from_webrtc_lidar
from dimos.utils.testing import SensorReplay


def test_init() -> None:
    lidar = SensorReplay("office_lidar")

    for raw_frame in itertools.islice(lidar.iterate(), 5):
        assert isinstance(raw_frame, dict)
        frame = pointcloud2_from_webrtc_lidar(raw_frame)
        assert isinstance(frame, PointCloud2)



# Copyright 2025-2026 Dimensional Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from abc import ABC, abstractmethod
import asyncio
from collections.abc import AsyncIterator, Callable
from contextlib import asynccontextmanager
from dataclasses import dataclass
import pickle
from typing import Any, Generic, TypeVar

MsgT = TypeVar("MsgT")
TopicT = TypeVar("TopicT")
EncodingT = TypeVar("EncodingT")


class PubSub(Generic[TopicT, MsgT], ABC):
    """Abstract base class for pub/sub implementations with sugar methods."""

    @abstractmethod
    def publish(self, topic: TopicT, message: MsgT) -> None:
        """Publish a message to a topic."""
        ...

    @abstractmethod
    def subscribe(
        self, topic: TopicT, callback: Callable[[MsgT, TopicT], None]
    ) -> Callable[[], None]:
        """Subscribe to a topic with a callback. returns unsubscribe function"""
        ...

    @dataclass(slots=True)
    class _Subscription:
        _bus: "PubSub[Any, Any]"
        _topic: Any
        _cb: Callable[[Any, Any], None]
        _unsubscribe_fn: Callable[[], None]

        def unsubscribe(self) -> None:
            self._unsubscribe_fn()

        # context-manager helper
        def __enter__(self):  # type: ignore[no-untyped-def]
            return self

        def __exit__(self, *exc) -> None:  # type: ignore[no-untyped-def]
            self.unsubscribe()

    # public helper: returns disposable object
    def sub(self, topic: TopicT, cb: Callable[[MsgT, TopicT], None]) -> "_Subscription":
        unsubscribe_fn = self.subscribe(topic, cb)
        return self._Subscription(self, topic, cb, unsubscribe_fn)

    # async iterator
    async def aiter(self, topic: TopicT, *, max_pending: int | None = None) -> AsyncIterator[MsgT]:
        q: asyncio.Queue[MsgT] = asyncio.Queue(maxsize=max_pending or 0)

        def _cb(msg: MsgT, topic: TopicT) -> None:
            q.put_nowait(msg)

        unsubscribe_fn = self.subscribe(topic, _cb)
        try:
            while True:
                yield await q.get()
        finally:
            unsubscribe_fn()

        # async context manager returning a queue

    @asynccontextmanager
    async def queue(self, topic: TopicT, *, max_pending: int | None = None):  # type: ignore[no-untyped-def]
        q: asyncio.Queue[MsgT] = asyncio.Queue(maxsize=max_pending or 0)

        def _queue_cb(msg: MsgT, topic: TopicT) -> None:
            q.put_nowait(msg)

        unsubscribe_fn = self.subscribe(topic, _queue_cb)
        try:
            yield q
        finally:
            unsubscribe_fn()


class PubSubEncoderMixin(Generic[TopicT, MsgT, EncodingT], ABC):
    """Mixin that encodes messages before publishing and decodes them after receiving.

    Usage: Just specify encoder and decoder as a subclass:

    class MyPubSubWithJSON(PubSubEncoderMixin, MyPubSub):
        def encoder(msg, topic):
            json.dumps(msg).encode('utf-8')
        def decoder(msg, topic):
            data: json.loads(data.decode('utf-8'))
    """

    @abstractmethod
    def encode(self, msg: MsgT, topic: TopicT) -> EncodingT: ...

    @abstractmethod
    def decode(self, msg: EncodingT, topic: TopicT) -> MsgT: ...

    def __init__(self, *args, **kwargs) -> None:  # type: ignore[no-untyped-def]
        super().__init__(*args, **kwargs)
        self._encode_callback_map: dict = {}  # type: ignore[type-arg]

    def publish(self, topic: TopicT, message: MsgT) -> None:
        """Encode the message and publish it."""
        if getattr(self, "_stop_event", None) is not None and self._stop_event.is_set():  # type: ignore[attr-defined]
            return
        encoded_message = self.encode(message, topic)
        if encoded_message is None:
            return
        super().publish(topic, encoded_message)  # type: ignore[misc]

    def subscribe(
        self, topic: TopicT, callback: Callable[[MsgT, TopicT], None]
    ) -> Callable[[], None]:
        """Subscribe with automatic decoding."""

        def wrapper_cb(encoded_data: EncodingT, topic: TopicT) -> None:
            decoded_message = self.decode(encoded_data, topic)
            callback(decoded_message, topic)

        return super().subscribe(topic, wrapper_cb)  # type: ignore[misc, no-any-return]


class PickleEncoderMixin(PubSubEncoderMixin[TopicT, MsgT, bytes]):
    def encode(self, msg: MsgT, *_: TopicT) -> bytes:  # type: ignore[return]
        try:
            return pickle.dumps(msg)
        except Exception as e:
            print("Pickle encoding error:", e)
            import traceback

            traceback.print_exc()
            print("Tried to pickle:", msg)

    def decode(self, msg: bytes, _: TopicT) -> MsgT:
        return pickle.loads(msg)  # type: ignore[no-any-return]



# Copyright 2025-2026 Dimensional Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import pickle

from dimos_lcm.sensor_msgs import CameraInfo

from dimos.msgs.sensor_msgs import Image, PointCloud2
from dimos.msgs.std_msgs import Header
from dimos.robot.unitree_webrtc.type.lidar import pointcloud2_from_webrtc_lidar
from dimos.robot.unitree_webrtc.type.odometry import Odometry

image_resize_factor = 1
originalwidth, originalheight = (1280, 720)


def camera_info() -> CameraInfo:
    fx, fy, cx, cy = list(
        map(
            lambda x: int(x / image_resize_factor),
            [819.553492, 820.646595, 625.284099, 336.808987],
        )
    )
    width, height = tuple(
        map(
            lambda x: int(x / image_resize_factor),
            [originalwidth, originalheight],
        )
    )

    # Camera matrix K (3x3)
    K = [fx, 0, cx, 0, fy, cy, 0, 0, 1]

    # No distortion coefficients for now
    D = [0.0, 0.0, 0.0, 0.0, 0.0]

    # Identity rotation matrix
    R = [1, 0, 0, 0, 1, 0, 0, 0, 1]

    # Projection matrix P (3x4)
    P = [fx, 0, cx, 0, 0, fy, cy, 0, 0, 0, 1, 0]

    base_msg = {
        "D_length": len(D),
        "height": height,
        "width": width,
        "distortion_model": "plumb_bob",
        "D": D,
        "K": K,
        "R": R,
        "P": P,
        "binning_x": 0,
        "binning_y": 0,
    }

    return CameraInfo(
        **base_msg,
        header=Header("camera_optical"),
    )


def transform_chain(odom_frame: Odometry) -> list:  # type: ignore[type-arg]
    from dimos.msgs.geometry_msgs import Quaternion, Transform, Vector3
    from dimos.protocol.tf import TF

    camera_link = Transform(
        translation=Vector3(0.3, 0.0, 0.0),
        rotation=Quaternion(0.0, 0.0, 0.0, 1.0),
        frame_id="base_link",
        child_frame_id="camera_link",
        ts=odom_frame.ts,
    )

    camera_optical = Transform(
        translation=Vector3(0.0, 0.0, 0.0),
        rotation=Quaternion(-0.5, 0.5, -0.5, 0.5),
        frame_id="camera_link",
        child_frame_id="camera_optical",
        ts=camera_link.ts,
    )

    tf = TF()
    tf.publish(
        Transform.from_pose("base_link", odom_frame),
        camera_link,
        camera_optical,
    )

    return tf  # type: ignore[return-value]


def broadcast(  # type: ignore[no-untyped-def]
    timestamp: float,
    lidar_frame: PointCloud2,
    video_frame: Image,
    odom_frame: Odometry,
    detections,
    annotations,
) -> None:
    from dimos_lcm.foxglove_msgs.ImageAnnotations import (
        ImageAnnotations,
    )

    from dimos.core import LCMTransport
    from dimos.msgs.geometry_msgs import PoseStamped

    lidar_transport = LCMTransport("/lidar", PointCloud2)  # type: ignore[var-annotated]
    odom_transport = LCMTransport("/odom", PoseStamped)  # type: ignore[var-annotated]
    video_transport = LCMTransport("/image", Image)  # type: ignore[var-annotated]
    camera_info_transport = LCMTransport("/camera_info", CameraInfo)  # type: ignore[var-annotated]

    lidar_transport.broadcast(None, lidar_frame)
    video_transport.broadcast(None, video_frame)
    odom_transport.broadcast(None, odom_frame)
    camera_info_transport.broadcast(None, camera_info())

    transform_chain(odom_frame)

    print(lidar_frame)
    print(video_frame)
    print(odom_frame)
    video_transport = LCMTransport("/image", Image)
    annotations_transport = LCMTransport("/annotations", ImageAnnotations)  # type: ignore[var-annotated]
    annotations_transport.broadcast(None, annotations)


def process_data():  # type: ignore[no-untyped-def]
    from dimos.msgs.sensor_msgs import Image
    from dimos.perception.detection.module2D import (  # type: ignore[attr-defined]
        Detection2DModule,
        build_imageannotations,
    )
    from dimos.robot.unitree_webrtc.type.odometry import Odometry
    from dimos.utils.data import get_data
    from dimos.utils.testing import TimedSensorReplay

    get_data("unitree_office_walk")
    target = 1751591272.9654856
    lidar_store = TimedSensorReplay(
        "unitree_office_walk/lidar", autocast=pointcloud2_from_webrtc_lidar
    )
    video_store = TimedSensorReplay("unitree_office_walk/video", autocast=Image.from_numpy)
    odom_store = TimedSensorReplay("unitree_office_walk/odom", autocast=Odometry.from_msg)

    def attach_frame_id(image: Image) -> Image:
        image.frame_id = "camera_optical"
        return image

    lidar_frame = lidar_store.find_closest(target, tolerance=1)
    video_frame = attach_frame_id(video_store.find_closest(target, tolerance=1))  # type: ignore[arg-type]
    odom_frame = odom_store.find_closest(target, tolerance=1)

    detector = Detection2DModule()
    detections = detector.detect(video_frame)  # type: ignore[attr-defined]
    annotations = build_imageannotations(detections)

    data = (target, lidar_frame, video_frame, odom_frame, detections, annotations)

    with open("filename.pkl", "wb") as file:
        pickle.dump(data, file)

    return data


def main() -> None:
    try:
        with open("filename.pkl", "rb") as file:
            data = pickle.load(file)
    except FileNotFoundError:
        print("Processing data and creating pickle file...")
        data = process_data()  # type: ignore[no-untyped-call]
    broadcast(*data)


main()



# Copyright 2025-2026 Dimensional Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
import numpy as np
import pytest

from dimos.types.vector import Vector


def test_vector_default_init() -> None:
    """Test that default initialization of Vector() has x,y,z components all zero."""
    v = Vector()
    assert v.x == 0.0
    assert v.y == 0.0
    assert v.z == 0.0
    assert v.dim == 0
    assert len(v.data) == 0
    assert v.to_list() == []
    assert v.is_zero()  # Empty vector should be considered zero


def test_vector_specific_init() -> None:
    """Test initialization with specific values."""
    # 2D vector
    v1 = Vector(1.0, 2.0)
    assert v1.x == 1.0
    assert v1.y == 2.0
    assert v1.z == 0.0
    assert v1.dim == 2

    # 3D vector
    v2 = Vector(3.0, 4.0, 5.0)
    assert v2.x == 3.0
    assert v2.y == 4.0
    assert v2.z == 5.0
    assert v2.dim == 3

    # From list
    v3 = Vector([6.0, 7.0, 8.0])
    assert v3.x == 6.0
    assert v3.y == 7.0
    assert v3.z == 8.0
    assert v3.dim == 3

    # From numpy array
    v4 = Vector(np.array([9.0, 10.0, 11.0]))
    assert v4.x == 9.0
    assert v4.y == 10.0
    assert v4.z == 11.0
    assert v4.dim == 3


def test_vector_addition() -> None:
    """Test vector addition."""
    v1 = Vector(1.0, 2.0, 3.0)
    v2 = Vector(4.0, 5.0, 6.0)

    v_add = v1 + v2
    assert v_add.x == 5.0
    assert v_add.y == 7.0
    assert v_add.z == 9.0


def test_vector_subtraction() -> None:
    """Test vector subtraction."""
    v1 = Vector(1.0, 2.0, 3.0)
    v2 = Vector(4.0, 5.0, 6.0)

    v_sub = v2 - v1
    assert v_sub.x == 3.0
    assert v_sub.y == 3.0
    assert v_sub.z == 3.0


def test_vector_scalar_multiplication() -> None:
    """Test vector multiplication by a scalar."""
    v1 = Vector(1.0, 2.0, 3.0)

    v_mul = v1 * 2.0
    assert v_mul.x == 2.0
    assert v_mul.y == 4.0
    assert v_mul.z == 6.0

    # Test right multiplication
    v_rmul = 2.0 * v1
    assert v_rmul.x == 2.0
    assert v_rmul.y == 4.0
    assert v_rmul.z == 6.0


def test_vector_scalar_division() -> None:
    """Test vector division by a scalar."""
    v2 = Vector(4.0, 5.0, 6.0)

    v_div = v2 / 2.0
    assert v_div.x == 2.0
    assert v_div.y == 2.5
    assert v_div.z == 3.0


def test_vector_dot_product() -> None:
    """Test vector dot product."""
    v1 = Vector(1.0, 2.0, 3.0)
    v2 = Vector(4.0, 5.0, 6.0)

    dot = v1.dot(v2)
    assert dot == 32.0


def test_vector_length() -> None:
    """Test vector length calculation."""
    # 2D vector with length 5
    v1 = Vector(3.0, 4.0)
    assert v1.length() == 5.0

    # 3D vector
    v2 = Vector(2.0, 3.0, 6.0)
    assert v2.length() == pytest.approx(7.0, 0.001)

    # Test length_squared
    assert v1.length_squared() == 25.0
    assert v2.length_squared() == 49.0


def test_vector_normalize() -> None:
    """Test vector normalization."""
    v = Vector(2.0, 3.0, 6.0)
    assert not v.is_zero()

    v_norm = v.normalize()
    length = v.length()
    expected_x = 2.0 / length
    expected_y = 3.0 / length
    expected_z = 6.0 / length

    assert np.isclose(v_norm.x, expected_x)
    assert np.isclose(v_norm.y, expected_y)
    assert np.isclose(v_norm.z, expected_z)
    assert np.isclose(v_norm.length(), 1.0)
    assert not v_norm.is_zero()

    # Test normalizing a zero vector
    v_zero = Vector(0.0, 0.0, 0.0)
    assert v_zero.is_zero()
    v_zero_norm = v_zero.normalize()
    assert v_zero_norm.x == 0.0
    assert v_zero_norm.y == 0.0
    assert v_zero_norm.z == 0.0
    assert v_zero_norm.is_zero()


def test_vector_to_2d() -> None:
    """Test conversion to 2D vector."""
    v = Vector(2.0, 3.0, 6.0)

    v_2d = v.to_2d()
    assert v_2d.x == 2.0
    assert v_2d.y == 3.0
    assert v_2d.z == 0.0
    assert v_2d.dim == 2

    # Already 2D vector
    v2 = Vector(4.0, 5.0)
    v2_2d = v2.to_2d()
    assert v2_2d.x == 4.0
    assert v2_2d.y == 5.0
    assert v2_2d.dim == 2


def test_vector_distance() -> None:
    """Test distance calculations between vectors."""
    v1 = Vector(1.0, 2.0, 3.0)
    v2 = Vector(4.0, 6.0, 8.0)

    # Distance
    dist = v1.distance(v2)
    expected_dist = np.sqrt(9.0 + 16.0 + 25.0)  # sqrt((4-1)² + (6-2)² + (8-3)²)
    assert dist == pytest.approx(expected_dist)

    # Distance squared
    dist_sq = v1.distance_squared(v2)
    assert dist_sq == 50.0  # 9 + 16 + 25


def test_vector_cross_product() -> None:
    """Test vector cross product."""
    v1 = Vector(1.0, 0.0, 0.0)  # Unit x vector
    v2 = Vector(0.0, 1.0, 0.0)  # Unit y vector

    # v1 × v2 should be unit z vector
    cross = v1.cross(v2)
    assert cross.x == 0.0
    assert cross.y == 0.0
    assert cross.z == 1.0

    # Test with more complex vectors
    a = Vector(2.0, 3.0, 4.0)
    b = Vector(5.0, 6.0, 7.0)
    c = a.cross(b)

    # Cross product manually calculated:
    # (3*7-4*6, 4*5-2*7, 2*6-3*5)
    assert c.x == -3.0
    assert c.y == 6.0
    assert c.z == -3.0

    # Test with 2D vectors (should raise error)
    v_2d = Vector(1.0, 2.0)
    with pytest.raises(ValueError):
        v_2d.cross(v2)


def test_vector_zeros() -> None:
    """Test Vector.zeros class method."""
    # 3D zero vector
    v_zeros = Vector.zeros(3)
    assert v_zeros.x == 0.0
    assert v_zeros.y == 0.0
    assert v_zeros.z == 0.0
    assert v_zeros.dim == 3
    assert v_zeros.is_zero()

    # 2D zero vector
    v_zeros_2d = Vector.zeros(2)
    assert v_zeros_2d.x == 0.0
    assert v_zeros_2d.y == 0.0
    assert v_zeros_2d.z == 0.0
    assert v_zeros_2d.dim == 2
    assert v_zeros_2d.is_zero()


def test_vector_ones() -> None:
    """Test Vector.ones class method."""
    # 3D ones vector
    v_ones = Vector.ones(3)
    assert v_ones.x == 1.0
    assert v_ones.y == 1.0
    assert v_ones.z == 1.0
    assert v_ones.dim == 3

    # 2D ones vector
    v_ones_2d = Vector.ones(2)
    assert v_ones_2d.x == 1.0
    assert v_ones_2d.y == 1.0
    assert v_ones_2d.z == 0.0
    assert v_ones_2d.dim == 2


def test_vector_conversion_methods() -> None:
    """Test vector conversion methods (to_list, to_tuple, to_numpy)."""
    v = Vector(1.0, 2.0, 3.0)

    # to_list
    assert v.to_list() == [1.0, 2.0, 3.0]

    # to_tuple
    assert v.to_tuple() == (1.0, 2.0, 3.0)

    # to_numpy
    np_array = v.to_numpy()
    assert isinstance(np_array, np.ndarray)
    assert np.array_equal(np_array, np.array([1.0, 2.0, 3.0]))


def test_vector_equality() -> None:
    """Test vector equality."""
    v1 = Vector(1, 2, 3)
    v2 = Vector(1, 2, 3)
    v3 = Vector(4, 5, 6)

    assert v1 == v2
    assert v1 != v3
    assert v1 != Vector(1, 2)  # Different dimensions
    assert v1 != Vector(1.1, 2, 3)  # Different values
    assert v1 != [1, 2, 3]


def test_vector_is_zero() -> None:
    """Test is_zero method for vectors."""
    # Default empty vector
    v0 = Vector()
    assert v0.is_zero()

    # Explicit zero vector
    v1 = Vector(0.0, 0.0, 0.0)
    assert v1.is_zero()

    # Zero vector with different dimensions
    v2 = Vector(0.0, 0.0)
    assert v2.is_zero()

    # Non-zero vectors
    v3 = Vector(1.0, 0.0, 0.0)
    assert not v3.is_zero()

    v4 = Vector(0.0, 2.0, 0.0)
    assert not v4.is_zero()

    v5 = Vector(0.0, 0.0, 3.0)
    assert not v5.is_zero()

    # Almost zero (within tolerance)
    v6 = Vector(1e-10, 1e-10, 1e-10)
    assert v6.is_zero()

    # Almost zero (outside tolerance)
    v7 = Vector(1e-6, 1e-6, 1e-6)
    assert not v7.is_zero()


def test_vector_bool_conversion():
    """Test boolean conversion of vectors."""
    # Zero vectors should be False
    v0 = Vector()
    assert not bool(v0)

    v1 = Vector(0.0, 0.0, 0.0)
    assert not bool(v1)

    # Almost zero vectors should be False
    v2 = Vector(1e-10, 1e-10, 1e-10)
    assert not bool(v2)

    # Non-zero vectors should be True
    v3 = Vector(1.0, 0.0, 0.0)
    assert bool(v3)

    v4 = Vector(0.0, 2.0, 0.0)
    assert bool(v4)

    v5 = Vector(0.0, 0.0, 3.0)
    assert bool(v5)

    # Direct use in if statements
    if v0:
        raise AssertionError("Zero vector should be False in boolean context")
    else:
        pass  # Expected path

    if v3:
        pass  # Expected path
    else:
        raise AssertionError("Non-zero vector should be True in boolean context")


def test_vector_add() -> None:
    """Test vector addition operator."""
    v1 = Vector(1.0, 2.0, 3.0)
    v2 = Vector(4.0, 5.0, 6.0)

    # Using __add__ method
    v_add = v1.__add__(v2)
    assert v_add.x == 5.0
    assert v_add.y == 7.0
    assert v_add.z == 9.0

    # Using + operator
    v_add_op = v1 + v2
    assert v_add_op.x == 5.0
    assert v_add_op.y == 7.0
    assert v_add_op.z == 9.0

    # Adding zero vector should return original vector
    v_zero = Vector.zeros(3)
    assert (v1 + v_zero) == v1


def test_vector_add_dim_mismatch() -> None:
    """Test vector addition operator."""
    v1 = Vector(1.0, 2.0)
    v2 = Vector(4.0, 5.0, 6.0)

    # Using + operator
    v1 + v2



# Copyright 2025-2026 Dimensional Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from queue import Queue

import cv2
from flask import Flask, Response, render_template
from reactivex import operators as ops
from reactivex.disposable import SingleAssignmentDisposable

from dimos.web.edge_io import EdgeIO


class FlaskServer(EdgeIO):
    def __init__(  # type: ignore[no-untyped-def]
        self,
        dev_name: str = "Flask Server",
        edge_type: str = "Bidirectional",
        port: int = 5555,
        **streams,
    ) -> None:
        super().__init__(dev_name, edge_type)
        self.app = Flask(__name__)
        self.port = port
        self.streams = streams
        self.active_streams = {}

        # Initialize shared stream references with ref_count
        for key in self.streams:
            if self.streams[key] is not None:
                # Apply share and ref_count to manage subscriptions
                self.active_streams[key] = self.streams[key].pipe(
                    ops.map(self.process_frame_flask), ops.share()
                )

        self.setup_routes()

    def process_frame_flask(self, frame):  # type: ignore[no-untyped-def]
        """Convert frame to JPEG format for streaming."""
        _, buffer = cv2.imencode(".jpg", frame)
        return buffer.tobytes()

    def setup_routes(self) -> None:
        @self.app.route("/")
        def index():  # type: ignore[no-untyped-def]
            stream_keys = list(self.streams.keys())  # Get the keys from the streams dictionary
            return render_template("index_flask.html", stream_keys=stream_keys)

        # Function to create a streaming response
        def stream_generator(key):  # type: ignore[no-untyped-def]
            def generate():  # type: ignore[no-untyped-def]
                frame_queue = Queue()  # type: ignore[var-annotated]
                disposable = SingleAssignmentDisposable()

                # Subscribe to the shared, ref-counted stream
                if key in self.active_streams:
                    disposable.disposable = self.active_streams[key].subscribe(
                        lambda frame: frame_queue.put(frame) if frame is not None else None,
                        lambda e: frame_queue.put(None),
                        lambda: frame_queue.put(None),
                    )

                try:
                    while True:
                        frame = frame_queue.get()
                        if frame is None:
                            break
                        yield (b"--frame\r\nContent-Type: image/jpeg\r\n\r\n" + frame + b"\r\n")
                finally:
                    disposable.dispose()

            return generate

        def make_response_generator(key):  # type: ignore[no-untyped-def]
            def response_generator():  # type: ignore[no-untyped-def]
                return Response(
                    stream_generator(key)(),  # type: ignore[no-untyped-call]
                    mimetype="multipart/x-mixed-replace; boundary=frame",
                )

            return response_generator

        # Dynamically adding routes using add_url_rule
        for key in self.streams:
            endpoint = f"video_feed_{key}"
            self.app.add_url_rule(
                f"/video_feed/{key}",
                endpoint,
                view_func=make_response_generator(key),  # type: ignore[no-untyped-call]
            )

    def run(self, host: str = "0.0.0.0", port: int = 5555, threaded: bool = True) -> None:
        self.port = port
        self.app.run(host=host, port=self.port, debug=False, threaded=threaded)



# Copyright 2025-2026 Dimensional Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import numpy as np
import pytest
from reactivex import operators as ops

from dimos.msgs.sensor_msgs.Image import Image, ImageFormat, sharpness_barrier
from dimos.utils.data import get_data
from dimos.utils.testing import TimedSensorReplay


@pytest.fixture
def img():
    image_file_path = get_data("cafe.jpg")
    return Image.from_file(str(image_file_path))


def test_file_load(img: Image) -> None:
    assert isinstance(img.data, np.ndarray)
    assert img.width == 1024
    assert img.height == 771
    assert img.channels == 3
    assert img.shape == (771, 1024, 3)
    assert img.data.dtype == np.uint8
    assert img.format == ImageFormat.BGR
    assert img.frame_id == ""
    assert isinstance(img.ts, float)
    assert img.ts > 0
    assert img.data.flags["C_CONTIGUOUS"]


def test_lcm_encode_decode(img: Image) -> None:
    binary_msg = img.lcm_encode()
    decoded_img = Image.lcm_decode(binary_msg)

    assert isinstance(decoded_img, Image)
    assert decoded_img is not img
    assert decoded_img == img


def test_rgb_bgr_conversion(img: Image) -> None:
    rgb = img.to_rgb()
    assert not rgb == img
    assert rgb.to_bgr() == img


def test_opencv_conversion(img: Image) -> None:
    ocv = img.to_opencv()
    decoded_img = Image.from_opencv(ocv)

    # artificially patch timestamp
    decoded_img.ts = img.ts
    assert decoded_img == img


@pytest.mark.tool
def test_sharpness_stream() -> None:
    get_data("unitree_office_walk")  # Preload data for testing
    video_store = TimedSensorReplay(
        "unitree_office_walk/video", autocast=lambda x: Image.from_numpy(x).to_rgb()
    )

    cnt = 0
    for image in video_store.iterate():
        cnt = cnt + 1
        print(image.sharpness)
        if cnt > 30:
            return


def test_sharpness_barrier() -> None:
    import time
    from unittest.mock import MagicMock

    # Create mock images with known sharpness values
    # This avoids loading real data from disk
    mock_images = []
    sharpness_values = [0.3711, 0.3241, 0.3067, 0.2583, 0.3665]  # Just 5 images for 1 window

    for i, sharp in enumerate(sharpness_values):
        img = MagicMock()
        img.sharpness = sharp
        img.ts = 1758912038.208 + i * 0.01  # Simulate timestamps
        mock_images.append(img)

    # Track what goes into windows and what comes out
    start_wall_time = None
    window_contents = []  # List of (wall_time, image)
    emitted_images = []

    def track_input(img):
        """Track all images going into sharpness_barrier with wall-clock time"""
        nonlocal start_wall_time
        wall_time = time.time()
        if start_wall_time is None:
            start_wall_time = wall_time
        relative_time = wall_time - start_wall_time
        window_contents.append((relative_time, img))
        return img

    def track_output(img) -> None:
        """Track what sharpness_barrier emits"""
        emitted_images.append(img)

    # Use 20Hz frequency (0.05s windows) for faster test
    # Emit images at 100Hz to get ~5 per window
    from reactivex import from_iterable, interval

    source = from_iterable(mock_images).pipe(
        ops.zip(interval(0.01)),  # 100Hz emission rate
        ops.map(lambda x: x[0]),  # Extract just the image
    )

    source.pipe(
        ops.do_action(track_input),  # Track inputs
        sharpness_barrier(20),  # 20Hz = 0.05s windows
        ops.do_action(track_output),  # Track outputs
    ).run()

    # Only need 0.08s for 1 full window at 20Hz plus buffer
    time.sleep(0.08)

    # Verify we got correct emissions (items span across 2 windows due to timing)
    # Items 1-4 arrive in first window (0-50ms), item 5 arrives in second window (50-100ms)
    assert len(emitted_images) == 2, (
        f"Expected exactly 2 emissions (one per window), got {len(emitted_images)}"
    )

    # Group inputs by wall-clock windows and verify we got the sharpest

    # Verify each window emitted the sharpest image from that window
    # First window (0-50ms): items 1-4
    assert emitted_images[0].sharpness == 0.3711  # Highest among first 4 items

    # Second window (50-100ms): only item 5
    assert emitted_images[1].sharpness == 0.3665  # Only item in second window



#!/usr/bin/env python3
# Copyright 2025-2026 Dimensional Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import time
from typing import Any

import numpy as np
from reactivex import Observable, create, disposable
import sounddevice as sd  # type: ignore[import-untyped]

from dimos.stream.audio.base import (
    AbstractAudioEmitter,
    AudioEvent,
)
from dimos.utils.logging_config import setup_logger

logger = setup_logger()


class SounddeviceAudioSource(AbstractAudioEmitter):
    """Audio source implementation using the sounddevice library."""

    def __init__(
        self,
        device_index: int | None = None,
        sample_rate: int = 16000,
        channels: int = 1,
        block_size: int = 1024,
        dtype: np.dtype = np.float32,  # type: ignore[assignment, type-arg]
    ) -> None:
        """
        Initialize SounddeviceAudioSource.

        Args:
            device_index: Audio device index (None for default)
            sample_rate: Audio sample rate in Hz
            channels: Number of audio channels (1=mono, 2=stereo)
            block_size: Number of samples per audio frame
            dtype: Data type for audio samples (np.float32 or np.int16)
        """
        self.device_index = device_index
        self.sample_rate = sample_rate
        self.channels = channels
        self.block_size = block_size
        self.dtype = dtype

        self._stream = None
        self._running = False

    def emit_audio(self) -> Observable:  # type: ignore[type-arg]
        """
        Create an observable that emits audio frames.

        Returns:
            Observable emitting AudioEvent objects
        """

        def on_subscribe(observer, scheduler):  # type: ignore[no-untyped-def]
            # Callback function to process audio data
            def audio_callback(indata, frames, time_info, status) -> None:  # type: ignore[no-untyped-def]
                if status:
                    logger.warning(f"Audio callback status: {status}")

                # Create audio event
                audio_event = AudioEvent(
                    data=indata.copy(),
                    sample_rate=self.sample_rate,
                    timestamp=time.time(),
                    channels=self.channels,
                )

                observer.on_next(audio_event)

            # Start the audio stream
            try:
                self._stream = sd.InputStream(
                    device=self.device_index,
                    samplerate=self.sample_rate,
                    channels=self.channels,
                    blocksize=self.block_size,
                    dtype=self.dtype,
                    callback=audio_callback,
                )
                self._stream.start()  # type: ignore[attr-defined]
                self._running = True

                logger.info(
                    f"Started audio capture: {self.sample_rate}Hz, "
                    f"{self.channels} channels, {self.block_size} samples per frame"
                )

            except Exception as e:
                logger.error(f"Error starting audio stream: {e}")
                observer.on_error(e)

            # Return a disposable to clean up resources
            def dispose() -> None:
                logger.info("Stopping audio capture")
                self._running = False
                if self._stream:
                    self._stream.stop()
                    self._stream.close()
                    self._stream = None

            return disposable.Disposable(dispose)

        return create(on_subscribe)

    def get_available_devices(self) -> list[dict[str, Any]]:
        """Get a list of available audio input devices."""
        return sd.query_devices()  # type: ignore[no-any-return]


if __name__ == "__main__":
    from dimos.stream.audio.node_volume_monitor import monitor
    from dimos.stream.audio.utils import keepalive

    monitor(SounddeviceAudioSource().emit_audio())
    keepalive()



# Copyright 2025-2026 Dimensional Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import pytest
import pytest_asyncio

from dimos.agents.agent import Agent
from dimos.core import start
from dimos.protocol.skill.test_coordinator import SkillContainerTest

system_prompt = (
    "Your name is Mr. Potato, potatoes are bad at math. Use a tools if asked to calculate"
)


@pytest.fixture(scope="session")
def dimos_cluster():
    """Session-scoped fixture to initialize dimos cluster once."""
    dimos = start(2)
    try:
        yield dimos
    finally:
        dimos.shutdown()


@pytest_asyncio.fixture
async def local():
    """Local context: both agent and testcontainer run locally"""
    testcontainer = SkillContainerTest()
    agent = Agent(system_prompt=system_prompt)
    try:
        yield agent, testcontainer
    except Exception as e:
        print(f"Error: {e}")
        import traceback

        traceback.print_exc()
        raise e
    finally:
        # Ensure cleanup happens while event loop is still active
        try:
            agent.stop()
        except Exception:
            pass
        try:
            testcontainer.stop()
        except Exception:
            pass


@pytest_asyncio.fixture
async def dask_mixed(dimos_cluster):
    """Dask context: testcontainer on dimos, agent local"""
    testcontainer = dimos_cluster.deploy(SkillContainerTest)
    agent = Agent(system_prompt=system_prompt)
    try:
        yield agent, testcontainer
    finally:
        try:
            agent.stop()
        except Exception:
            pass
        try:
            testcontainer.stop()
        except Exception:
            pass


@pytest_asyncio.fixture
async def dask_full(dimos_cluster):
    """Dask context: both agent and testcontainer deployed on dimos"""
    testcontainer = dimos_cluster.deploy(SkillContainerTest)
    agent = dimos_cluster.deploy(Agent, system_prompt=system_prompt)
    try:
        yield agent, testcontainer
    finally:
        try:
            agent.stop()
        except Exception:
            pass
        try:
            testcontainer.stop()
        except Exception:
            pass


@pytest_asyncio.fixture(params=["local", "dask_mixed", "dask_full"])
async def agent_context(request):
    """Parametrized fixture that runs tests with different agent configurations"""
    param = request.param

    if param == "local":
        testcontainer = SkillContainerTest()
        agent = Agent(system_prompt=system_prompt)
        try:
            yield agent, testcontainer
        finally:
            try:
                agent.stop()
            except Exception:
                pass
            try:
                testcontainer.stop()
            except Exception:
                pass
    elif param == "dask_mixed":
        dimos_cluster = request.getfixturevalue("dimos_cluster")
        testcontainer = dimos_cluster.deploy(SkillContainerTest)
        agent = Agent(system_prompt=system_prompt)
        try:
            yield agent, testcontainer
        finally:
            try:
                agent.stop()
            except Exception:
                pass
            try:
                testcontainer.stop()
            except Exception:
                pass
    elif param == "dask_full":
        dimos_cluster = request.getfixturevalue("dimos_cluster")
        testcontainer = dimos_cluster.deploy(SkillContainerTest)
        agent = dimos_cluster.deploy(Agent, system_prompt=system_prompt)
        try:
            yield agent, testcontainer
        finally:
            try:
                agent.stop()
            except Exception:
                pass
            try:
                testcontainer.stop()
            except Exception:
                pass


# @pytest.mark.timeout(40)
@pytest.mark.tool
@pytest.mark.asyncio
async def test_agent_init(agent_context) -> None:
    """Test agent initialization and basic functionality across different configurations"""
    agent, testcontainer = agent_context

    agent.register_skills(testcontainer)
    agent.start()

    # agent.run_implicit_skill("uptime_seconds")

    print("query agent")
    # When running locally, call the async method directly
    agent.query(
        "hi there, please tell me what's your name and current date, and how much is 124181112 + 124124?"
    )
    print("Agent loop finished, asking about camera")
    agent.query("tell me what you see on the camera?")

    # you can run skillspy and agentspy in parallel with this test for a better observation of what's happening



# Copyright 2025-2026 Dimensional Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""Helper utility functions for temporal memory."""

from typing import TYPE_CHECKING, Any

import numpy as np

if TYPE_CHECKING:
    from ..temporal_memory import Frame


def next_entity_id_hint(roster: Any) -> str:
    """Generate next entity ID based on existing roster (e.g., E1, E2, E3...)."""
    if not isinstance(roster, list):
        return "E1"
    max_n = 0
    for e in roster:
        if not isinstance(e, dict):
            continue
        eid = e.get("id")
        if isinstance(eid, str) and eid.startswith("E"):
            tail = eid[1:]
            if tail.isdigit():
                max_n = max(max_n, int(tail))
    return f"E{max_n + 1}"


def clamp_text(text: str, max_chars: int) -> str:
    """Clamp text to maximum characters."""
    if len(text) <= max_chars:
        return text
    return text[:max_chars] + "..."


def format_timestamp(seconds: float) -> str:
    """Format seconds as MM:SS.mmm timestamp string."""
    m = int(seconds // 60)
    s = seconds - 60 * m
    return f"{m:02d}:{s:06.3f}"


def is_scene_stale(frames: list["Frame"], stale_threshold: float = 5.0) -> bool:
    """Check if scene hasn't changed meaningfully between first and last frame.

    Args:
        frames: List of frames to check
        stale_threshold: Threshold for mean pixel difference (default: 5.0)

    Returns:
        True if scene is stale (hasn't changed enough), False otherwise
    """
    if len(frames) < 2:
        return False
    first_img = frames[0].image
    last_img = frames[-1].image
    if first_img is None or last_img is None:
        return False
    if not hasattr(first_img, "data") or not hasattr(last_img, "data"):
        return False
    diff = np.abs(first_img.data.astype(float) - last_img.data.astype(float))
    return bool(diff.mean() < stale_threshold)



# Copyright 2025-2026 Dimensional Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""TF Rerun Module - Snapshot TF visualization in Rerun.

This module polls the TF buffer at a configurable rate and logs the latest
transform for each edge to Rerun. This provides stable, rate-limited TF
visualization without subscribing to the /tf transport from here.

Usage:
    # In blueprints:
    from dimos.dashboard.tf_rerun_module import tf_rerun

    def my_robot():
        return (
            robot_connection()
            + tf_rerun()  # Add TF visualization
            + other_modules()
        )
"""

from collections.abc import Sequence
import threading
import time
from typing import Any, cast

import rerun as rr

from dimos.core import Module, rpc
from dimos.core.blueprints import ModuleBlueprintSet, autoconnect
from dimos.core.global_config import GlobalConfig
from dimos.dashboard.rerun_init import connect_rerun
from dimos.dashboard.rerun_scene_wiring import rerun_scene_wiring
from dimos.utils.logging_config import setup_logger

logger = setup_logger()


class TFRerunModule(Module):
    """Polls TF buffer and logs snapshot transforms to Rerun.

    This module automatically visualizes the TF tree in Rerun by:
    - Using `self.tf` (the system TF service) to maintain the TF buffer
    - Polling at a configurable rate and logging the latest transform per edge
    """

    _global_config: GlobalConfig
    _poll_thread: threading.Thread | None = None
    _stop_event: threading.Event | None = None
    _poll_hz: float
    _last_ts_by_edge: dict[tuple[str, str], float]

    def __init__(
        self,
        global_config: GlobalConfig | None = None,
        poll_hz: float = 30.0,
        **kwargs: Any,
    ) -> None:
        """Initialize TFRerunModule.

        Args:
            global_config: Optional global configuration for viewer backend settings
            **kwargs: Additional arguments passed to parent Module
        """
        super().__init__(**kwargs)
        self._global_config = global_config or GlobalConfig()
        self._poll_hz = poll_hz
        self._last_ts_by_edge = {}

    @rpc
    def start(self) -> None:
        """Start the TF visualization module."""
        super().start()

        # Only connect if Rerun backend is selected
        if self._global_config.viewer_backend.startswith("rerun"):
            connect_rerun(global_config=self._global_config)

            # Ensure TF transport is started so its internal subscription populates the buffer.
            self.tf.start(sub=True)

            self._stop_event = threading.Event()
            self._poll_thread = threading.Thread(target=self._poll_loop, daemon=True)
            self._poll_thread.start()
            logger.info("TFRerunModule: started TF snapshot polling", poll_hz=self._poll_hz)

    def _poll_loop(self) -> None:
        assert self._stop_event is not None
        period_s = 1.0 / max(self._poll_hz, 0.1)

        while not self._stop_event.is_set():
            # Snapshot keys to avoid concurrent modification while TF buffer updates.
            items = list(self.tf.buffers.items())  # type: ignore[attr-defined]
            for (parent, child), buffer in items:
                latest = buffer.get()
                if latest is None:
                    continue
                last_ts = self._last_ts_by_edge.get((parent, child))
                if last_ts is not None and latest.ts == last_ts:
                    continue

                # Log under `world/tf/...` so it is visible under the default 3D view origin.
                rr.log(f"world/tf/{child}", latest.to_rerun())  # type: ignore[no-untyped-call]
                self._last_ts_by_edge[(parent, child)] = latest.ts

            time.sleep(period_s)

    @rpc
    def stop(self) -> None:
        """Stop the TF visualization module and cleanup LCM subscription."""
        if self._stop_event is not None:
            self._stop_event.set()
            self._stop_event = None

        if self._poll_thread is not None and self._poll_thread.is_alive():
            self._poll_thread.join(timeout=1.0)
        self._poll_thread = None

        super().stop()


def tf_rerun(
    *,
    poll_hz: float = 30.0,
    scene: bool = True,
    # Scene wiring kwargs (only used if scene=True)
    world_entity: str = "world",
    robot_entity: str = "world/robot",
    robot_axes_entity: str = "world/robot/axes",
    world_frame: str = "world",
    robot_frame: str = "base_link",
    urdf_path: str | None = None,
    axes_size: float | None = 0.5,
    cameras: Sequence[tuple[str, str, Any]] = (),
    camera_rgb_suffix: str = "rgb",
) -> ModuleBlueprintSet:
    """Convenience blueprint: TF snapshot polling + (optional) static scene wiring.

    - TF visualization stays in `TFRerunModule` (poll TF buffer, log to `world/tf/*`).
    - Scene wiring is handled by `RerunSceneWiringModule` (view coords, attachments, URDF, pinholes).
    """
    tf_bp = cast("ModuleBlueprintSet", TFRerunModule.blueprint(poll_hz=poll_hz))
    if not scene:
        return tf_bp

    scene_bp = cast(
        "ModuleBlueprintSet",
        rerun_scene_wiring(
            world_entity=world_entity,
            robot_entity=robot_entity,
            robot_axes_entity=robot_axes_entity,
            world_frame=world_frame,
            robot_frame=robot_frame,
            urdf_path=urdf_path,
            axes_size=axes_size,
            cameras=cameras,
            camera_rgb_suffix=camera_rgb_suffix,
        ),
    )

    return autoconnect(tf_bp, scene_bp)



# Copyright 2025-2026 Dimensional Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from collections import deque

import numpy as np

from dimos.msgs.geometry_msgs import Vector3, VectorLike
from dimos.msgs.nav_msgs import CostValues, OccupancyGrid


def find_safe_goal(
    costmap: OccupancyGrid,
    goal: VectorLike,
    algorithm: str = "bfs",
    cost_threshold: int = 50,
    min_clearance: float = 0.3,
    max_search_distance: float = 5.0,
    connectivity_check_radius: int = 3,
) -> Vector3 | None:
    """
    Find a safe goal position when the original goal is in collision or too close to obstacles.

    Args:
        costmap: The occupancy grid/costmap
        goal: Original goal position in world coordinates
        algorithm: Algorithm to use ("bfs", "spiral", "voronoi", "gradient_descent")
        cost_threshold: Maximum acceptable cost for a safe position (default: 50)
        min_clearance: Minimum clearance from obstacles in meters (default: 0.3m)
        max_search_distance: Maximum distance to search from original goal in meters (default: 5.0m)
        connectivity_check_radius: Radius in cells to check for connectivity (default: 3)

    Returns:
        Safe goal position in world coordinates, or None if no safe position found
    """

    if algorithm == "bfs":
        return _find_safe_goal_bfs(
            costmap,
            goal,
            cost_threshold,
            min_clearance,
            max_search_distance,
            connectivity_check_radius,
        )
    elif algorithm == "bfs_contiguous":
        return _find_safe_goal_bfs_contiguous(
            costmap,
            goal,
            cost_threshold,
            min_clearance,
            max_search_distance,
            connectivity_check_radius,
        )
    else:
        raise ValueError(f"Unknown algorithm: {algorithm}")


def _find_safe_goal_bfs(
    costmap: OccupancyGrid,
    goal: VectorLike,
    cost_threshold: int,
    min_clearance: float,
    max_search_distance: float,
    connectivity_check_radius: int,
) -> Vector3 | None:
    """
    BFS-based search for nearest safe goal position.
    This guarantees finding the closest valid position.

    Pros:
    - Guarantees finding the closest safe position
    - Can check connectivity to avoid isolated spots
    - Efficient for small to medium search areas

    Cons:
    - Can be slower for large search areas
    - Memory usage scales with search area
    """

    # Convert goal to grid coordinates
    goal_grid = costmap.world_to_grid(goal)
    gx, gy = int(goal_grid.x), int(goal_grid.y)

    # Convert distances to grid cells
    clearance_cells = int(np.ceil(min_clearance / costmap.resolution))
    max_search_cells = int(np.ceil(max_search_distance / costmap.resolution))

    # BFS queue and visited set
    queue = deque([(gx, gy, 0)])
    visited = set([(gx, gy)])

    # 8-connected neighbors
    neighbors = [(0, 1), (1, 0), (0, -1), (-1, 0), (1, 1), (1, -1), (-1, 1), (-1, -1)]

    while queue:
        x, y, dist = queue.popleft()

        # Check if we've exceeded max search distance
        if dist > max_search_cells:
            break

        # Check if position is valid
        if _is_position_safe(
            costmap, x, y, cost_threshold, clearance_cells, connectivity_check_radius
        ):
            # Convert back to world coordinates
            return costmap.grid_to_world((x, y))

        # Add neighbors to queue
        for dx, dy in neighbors:
            nx, ny = x + dx, y + dy

            # Check bounds
            if 0 <= nx < costmap.width and 0 <= ny < costmap.height:
                if (nx, ny) not in visited:
                    visited.add((nx, ny))
                    queue.append((nx, ny, dist + 1))

    return None


def _find_safe_goal_bfs_contiguous(
    costmap: OccupancyGrid,
    goal: VectorLike,
    cost_threshold: int,
    min_clearance: float,
    max_search_distance: float,
    connectivity_check_radius: int,
) -> Vector3 | None:
    """
    BFS-based search for nearest safe goal position, only following passable cells.
    Unlike regular BFS, this only expands through cells with occupancy < 100,
    ensuring the path doesn't cross through impassable obstacles.

    Pros:
    - Guarantees finding the closest safe position reachable without crossing obstacles
    - Ensures connectivity to the goal through passable space
    - Good for finding safe positions in the same "room" or connected area

    Cons:
    - May not find nearby safe spots if they're on the other side of a wall
    - Slightly slower than regular BFS due to additional checks
    """

    # Convert goal to grid coordinates
    goal_grid = costmap.world_to_grid(goal)
    gx, gy = int(goal_grid.x), int(goal_grid.y)

    # Convert distances to grid cells
    clearance_cells = int(np.ceil(min_clearance / costmap.resolution))
    max_search_cells = int(np.ceil(max_search_distance / costmap.resolution))

    # BFS queue and visited set
    queue = deque([(gx, gy, 0)])
    visited = set([(gx, gy)])

    # 8-connected neighbors
    neighbors = [(0, 1), (1, 0), (0, -1), (-1, 0), (1, 1), (1, -1), (-1, 1), (-1, -1)]

    while queue:
        x, y, dist = queue.popleft()

        # Check if we've exceeded max search distance
        if dist > max_search_cells:
            break

        # Check if position is valid
        if _is_position_safe(
            costmap, x, y, cost_threshold, clearance_cells, connectivity_check_radius
        ):
            # Convert back to world coordinates
            return costmap.grid_to_world((x, y))

        # Add neighbors to queue
        for dx, dy in neighbors:
            nx, ny = x + dx, y + dy

            # Check bounds
            if 0 <= nx < costmap.width and 0 <= ny < costmap.height:
                if (nx, ny) not in visited:
                    # Only expand through passable cells (occupancy < 100)
                    if costmap.grid[ny, nx] < 100:
                        visited.add((nx, ny))
                        queue.append((nx, ny, dist + 1))

    return None


def _is_position_safe(
    costmap: OccupancyGrid,
    x: int,
    y: int,
    cost_threshold: int,
    clearance_cells: int,
    connectivity_check_radius: int,
) -> bool:
    """
    Check if a position is safe based on multiple criteria.

    Args:
        costmap: The occupancy grid
        x, y: Grid coordinates to check
        cost_threshold: Maximum acceptable cost
        clearance_cells: Minimum clearance in cells
        connectivity_check_radius: Radius to check for connectivity

    Returns:
        True if position is safe, False otherwise
    """

    # Check bounds first
    if not (0 <= x < costmap.width and 0 <= y < costmap.height):
        return False

    # Check if position itself is free
    if costmap.grid[y, x] >= cost_threshold or costmap.grid[y, x] == CostValues.UNKNOWN:
        return False

    # Check clearance around position
    for dy in range(-clearance_cells, clearance_cells + 1):
        for dx in range(-clearance_cells, clearance_cells + 1):
            nx, ny = x + dx, y + dy
            if 0 <= nx < costmap.width and 0 <= ny < costmap.height:
                # Check if within circular clearance
                if dx * dx + dy * dy <= clearance_cells * clearance_cells:
                    if costmap.grid[ny, nx] >= cost_threshold:
                        return False

    # Check connectivity (not surrounded by obstacles)
    # Count free neighbors in a larger radius
    free_count = 0
    total_count = 0

    for dy in range(-connectivity_check_radius, connectivity_check_radius + 1):
        for dx in range(-connectivity_check_radius, connectivity_check_radius + 1):
            if dx == 0 and dy == 0:
                continue

            nx, ny = x + dx, y + dy
            if 0 <= nx < costmap.width and 0 <= ny < costmap.height:
                total_count += 1
                if (
                    costmap.grid[ny, nx] < cost_threshold
                    and costmap.grid[ny, nx] != CostValues.UNKNOWN
                ):
                    free_count += 1

    # Require at least 50% of neighbors to be free (not surrounded)
    if total_count > 0 and free_count < total_count * 0.5:
        return False

    return True
